"""
NIMPA: functions for neuro image processing and analysis
including partial volume correction (PVC) and ROI extraction and analysis.
"""
import logging
import multiprocessing
import os
import re
import sys
from glob import glob
from subprocess import run
from textwrap import dedent
from warnings import warn

import numpy as np
import scipy.ndimage as ndi
from pkg_resources import resource_filename
from tqdm.auto import trange

from . import imio, regseg

try:
    # GPU routines if compiled
    import cuvec as cu

    from . import improc
except ImportError:
    cu = None
    improc = None
sitk_flag = True
try:
    import SimpleITK as sitk
except ImportError:
    sitk_flag = False

log = logging.getLogger(__name__)

# possible extentions for DICOM files
dcmext = ('dcm', 'DCM', 'ima', 'IMA')
niiext = ('nii.gz', 'nii', 'img', 'hdr')


# ----------------------------------------------------------------------
def num(s):
    '''Converts the string to a float or integer number.'''
    try:
        return int(s)
    except ValueError:
        return float(s)


# ----------------------------------------------------------------------
def psf_gaussian(vx_size=(1, 1, 1), fwhm=(6, 5, 5), hradius=8):
    '''
    Separable kernels for Gaussian convolution executed on the GPU device
    The output kernels are in this order: z, y, x
    '''

    # if voxel size is given as scalar, interpret it as an isotropic
    # voxel size.
    if isinstance(vx_size, (float, int)):
        vx_size = [vx_size, vx_size, vx_size]

    # the same for the Gaussian kernel
    if isinstance(fwhm, (float, int)):
        fwhm = [fwhm, fwhm, fwhm]

    # avoid zeros in FWHM
    fwhm = [x + 1e-3 * (x <= 0) for x in fwhm]

    xSig = (fwhm[2] / vx_size[2]) / (2 * (2 * np.log(2))**.5)
    ySig = (fwhm[1] / vx_size[1]) / (2 * (2 * np.log(2))**.5)
    zSig = (fwhm[0] / vx_size[0]) / (2 * (2 * np.log(2))**.5)

    # get the separated kernels
    x = np.arange(-hradius, hradius + 1)
    xKrnl = np.exp(-(x**2 / (2 * xSig**2)))
    yKrnl = np.exp(-(x**2 / (2 * ySig**2)))
    zKrnl = np.exp(-(x**2 / (2 * zSig**2)))

    # normalise kernels
    xKrnl /= np.sum(xKrnl)
    yKrnl /= np.sum(yKrnl)
    zKrnl /= np.sum(zKrnl)

    krnl = np.array([zKrnl, yKrnl, xKrnl], dtype=np.float32)

    # for checking if the normalisation worked
    # np.prod( np.sum(krnl,axis=1) )

    # return all kernels together
    return krnl


# ----------------------------------------------------------------------
def psf_measured(scanner='mmr', scale=1):
    if scanner == 'mmr':
        # file name for the mMR's PSF and chosen scale
        fdat = resource_filename("niftypet.nimpa", f"auxdata/PSF-17_scl-{scale:d}.npy")
        # transaxial and axial PSF
        Hxy, Hz = np.load(fdat)
        return np.array([Hz, Hxy, Hxy], dtype=np.float32)
    raise NameError(f'Unsupported scanner ({scanner}):'
                    ' only Siemens mMR (mmr) is currently supported')


# ----------------------------------------------------------------------
def conv_separable(vol, knl, dev_id=0):
    """
    Args:
      vol(ndarray): Can be any number of dimensions `ndim`
        (GPU requires `ndim <= 3`).
      knl(ndarray): `ndim` x `width` separable kernel
        (GPU requires `width <= 17`).
      dev_id(int or bool): GPU device ID to try [default: 0].
        Set to `False` to force CPU fallback.
    """
    assert vol.ndim == len(knl)
    assert knl.ndim == 2
    if len(knl) > 3 or knl.shape[1] > 17:
        log.warning("kernel larger than 3 x 17 not supported on GPU")
        dev_id = False
    if improc is not None and dev_id is not False:
        log.debug("GPU conv")
        pad = 3 - len(knl)        # <3 dims
        k_pad = 17 - knl.shape[1] # kernel width < 17
        if pad or k_pad:
            knl = np.pad(knl, [(0, pad), (k_pad // 2, k_pad//2 + (k_pad%2))])
            if pad:
                knl[-pad:, 17 // 2] = 1
                vol = vol.reshape(vol.shape + (1,) * pad)
        src = cu.asarray(vol, dtype='float32')
        knl = cu.asarray(knl, dtype='float32')
        dst = improc.convolve(src.cuvec, knl.cuvec, dev_id=dev_id, log=log.getEffectiveLevel())
        res = cu.asarray(dst, dtype=vol.dtype)
        return res[(slice(0, None),) * (res.ndim - pad) + (-1,) * pad] if pad else res
    else:
        log.debug("CPU conv")
        for dim in range(len(knl)):
            h = knl[dim].reshape((1,) * dim + (-1,) + (1,) * (len(knl) - dim - 1))
            vol = ndi.convolve(vol, h, mode='constant', cval=0.)
        return vol


def nlm(img, ref, sigma=1, half_width=4, output=None, dev_id=0):
    """
    3D Non-local means (NLM) guided filter.
    Args:
      img(3darray): input image to be filtered.
      ref(3darray): reference (guidance) image.
      sigma(float): NLM parameter.
      half_width(int): neighbourhood half-width.
      output(CuVec): pre-existing output memory.
    Reference: https://doi.org/10.1109/CVPR.2005.38
    """
    img = cu.asarray(img, 'float32')
    ref = cu.asarray(ref, 'float32')
    if img.shape != ref.shape:
        raise IndexError(f"{img.shape} and {ref.shape} don't match")
    if img.ndim != 3:
        raise IndexError(f"must be 3D: got {img.ndim}D")
    kwargs = {
        'sigma': sigma, 'half_width': half_width, 'dev_id': dev_id, 'log': log.getEffectiveLevel()}
    if output is not None:
        if not isinstance(output, cu.CuVec):
            raise TypeError("output must be a CuVec")
        if np.dtype(output.dtype) != np.dtype('float32'):
            raise TypeError(f"output must be float32: got {output.dtype}")
        if output.shape != img.shape:
            raise IndexError("output shape doesn't match")
        output = cu.asarray(output).cuvec
    return cu.asarray(improc.nlm(img.cuvec, ref.cuvec, **kwargs))


# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
# I M A G E   S M O O T H I N G
# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>


def imsmooth(fim, fwhm=4, psf=None, voxsize=None, fout='', output='image', gpu=None, dev_id=0,
             Cnt=None):
    '''
    Smooth image using Gaussian filter with either the PSF or FWHM given
    as an option.  By default FWHM = 4 is used with voxel size assumed 1 mm.
    Arguments:
    - fim:  can be a NIfTI image file or Numpy array
    - fwhm: the width at half max of the Gaussian kernel (z,y,x)
    - psf:  the point spread function for each direction (z,y,x) given as a
            Numpy matrix of 3x17 and used on the GPU as separable kernel
    - voxsize: size of the voxel (can be in mm or cm)
    - fout: the output file path
    - output: can be image as Numpy array or file or both.
    - dev_id: the ID of the CUDA device to try to use for computation
      (set to `False` to force disable GPU)
    - gpu: ignored
    '''
    if gpu is not None:
        warn("gpu is automatic", DeprecationWarning, stacklevel=2)
    if Cnt is not None and 'DEVID' in Cnt:
        dev_id = Cnt['DEVID']

    isfile = False
    if isinstance(fim, str) and os.path.isfile(fim):
        isfile = True
        imd = imio.getnii(fim, output='all')
        im = imd['im']
        voxsize = imd['voxsize']
        affine = imd['affine']
    elif isinstance(fim, dict) and 'voxsize' in fim:
        im = imd['im']
        voxsize = imd['voxsize']
        affine = imd['affine']
    elif isinstance(fim, (np.ndarray, np.generic)):
        im = fim
    else:
        raise ValueError(
            "incorrect image input.\nNIfTI file path, dictionary or Numpy array are accepted.")

    if voxsize is None and Cnt is not None and 'SO_VXZ' in Cnt:
        voxsize = [Cnt['SO_VXZ'], Cnt['SO_VXY'], Cnt['SO_VXX']]
    elif voxsize is None and Cnt is None:
        raise ValueError('the correct voxel size has to be provided')

    if psf is None:
        psf = psf_gaussian(vx_size=voxsize, fwhm=fwhm)
    imsmo = conv_separable(im, psf, dev_id=dev_id)

    # output dictionary
    dctout = {}
    dctout['im'] = imsmo
    dctout['fwhm'] = fwhm

    if isfile and fout == '':
        if fim.endswith('.nii.gz'):
            fout = fim.split('.nii.gz')[0] + '_smo' + str(fwhm).replace('.', '-') + '.nii.gz'
        else:
            fout = os.path.splitext(fim)[0] + '_smo' + str(fwhm).replace(
                '.', '-') + os.path.splitext(fim)[1]

    if isfile:
        dctout['affine'] = affine
        dctout['fim'] = fout

        imio.array2nii(
            imsmo, affine, fout, trnsp=(imd['transpose'].index(0), imd['transpose'].index(1),
                                        imd['transpose'].index(2)), flip=imd['flip'])

    if output == 'all':
        return dctout
    elif output == 'image':
        return imsmo
    elif output == 'file':
        return fout
    else:
        return None


# ==================================================================================================
# > get projections along 3D image axes
def im_project3(im):
    ''' project intensities on the 3 axes x, y ,z.
    '''

    if isinstance(im, str):
        img = imio.getnii(im)
    elif isinstance(im, (np.ndarray, np.generic)):
        img = im
    else:
        raise ValueError('unrecognised image input')

    qx = np.sum(img, axis=(0, 1))
    qy = np.sum(img, axis=(0, 2))
    qz = np.sum(img, axis=(1, 2))

    return qx, qy, qz


# <><><><><><><><><><><><><><><><><><><><><><><><><><><><>
# FUNCTIONS: TRIM &  PARTIAL VOLUME EFFECTS AND CORRECTION
# <><><><><><><><><><><><><><><><><><><><><><><><><><><><>


def imtrimup(fims, refim='', affine=None, scale=2, divdim=8**2, fmax=0.05, int_order=0, outpath='',
             fname='', fcomment='', fcomment_pfx='', store_avg=False, store_img_intrmd=False,
             store_img=False, imdtype=np.float32, memlim=False, verbose=False, Cnt=None):
    '''
    Trim and upsample PET image(s), e.g., for GPU execution,
    PVC correction, ROI sampling, etc.
    The input images 'fims' can be passed in multiple ways:
    1. as a string of the folder containing NIfTI files
    2. as a string of a NIfTI file path (this way a 4D image can be loaded).
    3. as a list of NIfTI file paths.
    4. as a 3D or 4D image
    Parameters:
    -----------
    refim:  Path to the reference image, which was already trimmed.
            Needs the trimming parameters stored in the NIfTI header in 'descrip'.
    affine: affine matrix, 4x4, for all the input images in case when images
            passed as a matrix (all have to have the same shape and data type)
    scale:  the scaling factor for upsampling has to be greater than 1.  It can be a scalar or
            a vector/tuple with elements for each image dimension.
    divdim: image divisor, i.e., the dimensions are a multiple of this number (default 64)
    int_order: interpolation order (0-nearest neighbour, 1-linear, as in scipy)
    fmax: fraction of the max image value used for finding image borders for trimming
    outpath: output folder path
    fname: file name when image given as a numpy matrix
    fcomment: part of the name of the output file, left for the user as a comment
    fcomment_pfx: part of the name of the output file at the start (prefix)
    store_img_intrmd: stores intermediate images with suffix '_i'
    store_avg: stores the average image (if multiple images are given)
    imdtype: data type for output images
    memlim: Ture for cases when memory is limited and takes more processing time instead.
    verbose: verbose mode [True/False]
    '''
    if Cnt is None:
        Cnt = {}

    # case when input folder is given
    if isinstance(fims, str) and os.path.isdir(fims):
        # list of input images (e.g., PET)
        fimlist = [os.path.join(fims, f) for f in os.listdir(fims) if f.endswith(niiext)]
        imdic = imio.niisort(fimlist, memlim=memlim)
        if not (imdic['N'] > 50 and memlim):
            imin = imdic['im']
        imshape = imdic['shape']
        affine = imdic['affine']
        fldrin = fims
        fnms = [os.path.basename(f).split('.nii')[0] for f in imdic['files'] if f is not None]
        # number of images/frames
        Nim = imdic['N']
        using_multiple_files = True

    # case when input file is a 3D or 4D NIfTI image
    elif isinstance(fims, str) and os.path.isfile(fims) and fims.endswith(niiext):
        imdic = imio.getnii(fims, output='all')
        imin = imdic['im']
        if imin.ndim == 3:
            imin.shape = (1, imin.shape[0], imin.shape[1], imin.shape[2])
        imdtype = imdic['dtype']
        imshape = imdic['shape'][-3:]
        affine = imdic['affine']
        fldrin = os.path.dirname(fims)
        fnms = imin.shape[0] * [os.path.basename(fims).split('.nii')[0]]
        # number of images/frames
        Nim = imin.shape[0]

    # case when a list of input files is given
    elif isinstance(fims, list) and all(map(os.path.isfile, fims)):
        imdic = imio.niisort(fims, memlim=memlim)
        if not (imdic['N'] > 50 and memlim):
            imin = imdic['im']
        imshape = imdic['shape']
        affine = imdic['affine']
        imdtype = imdic['dtype']
        fldrin = os.path.dirname(fims[0])
        fnms = [os.path.basename(f).split('.nii')[0] for f in imdic['files']]
        # number of images/frames
        Nim = imdic['N']
        using_multiple_files = True

    # case when an array [#frames, zdim, ydim, xdim].  Can be 3D or 4D
    elif isinstance(fims, (np.ndarray, np.generic)) and (fims.ndim == 4 or fims.ndim == 3):
        # check image affine
        if affine.shape != (4, 4):
            raise ValueError('Affine should be a 4x4 array.')
        # create a copy to avoid mutation when only one image (3D)
        imin = np.copy(fims)
        if fims.ndim == 3:
            imin.shape = (1, imin.shape[0], imin.shape[1], imin.shape[2])
        imshape = imin.shape[-3:]
        fldrin = os.path.join(os.path.expanduser('~'), 'NIMPA_output')
        if fname == '':
            fnms = imin.shape[0] * ['NIMPA']
        else:
            fnms = imin.shape[0] * [fname]
        # number of images/frames
        Nim = imin.shape[0]

    else:
        raise TypeError('Wrong data type input.')

    # > if using reference trimmed image:
    ref_flag = False
    if os.path.isfile(refim):
        refdct = imio.getnii(refim, output='all')
        nii_descrp = refdct['hdr']['descrip'].item().decode('utf-8')
        if 'trim' in nii_descrp:
            # > find all the numbers (int and floats)
            parstr = re.findall(r'-*\d+\.*\d*', nii_descrp)
            try:
                ix0, ix1, iy0, iy1, iz0, scale0, scale1, scale2, fmax = (num(s) for s in parstr)
                scale = [scale0, scale1, scale2]
            except ValueError:
                ix0, ix1, iy0, iy1, iz0, scale, fmax = (num(s) for s in parstr)
                scale = [scale, scale, scale]
            except Exception:
                log.error('the reference file does not have the trimming information.')

            ref_flag = True
            log.info(' using the trimming parameters of the reference image.')
        else:
            log.warning(' the reference image does not contain trimming info--using default.')

    # ------------------------------------------------------
    # store images in this folder
    if outpath == '':
        petudir = os.path.join(fldrin, 'trimmed')
    else:
        petudir = os.path.join(outpath, 'trimmed')
    imio.create_dir(petudir)
    # ------------------------------------------------------

    # ------------------------------------------------------
    # scale is preferred to be integer
    try:
        scale = np.int8(scale)
    except ValueError:
        raise ValueError('scale has to be an integer or array of integers.')

    # > check if scale is given as scalar of array.
    # > if scalar, change it to an array.
    if isinstance(scale, np.int8):
        scale = scale * np.ones(3, dtype=np.int8)

    # > scale factor as the inverse of scale
    sf = 1 / np.float64(scale)
    log.debug(' upsampling scale {}, giving resolution scale factor {} for {} images.'.format(
        scale, sf, Nim))
    # ------------------------------------------------------

    # ------------------------------------------------------
    # scaled input image and get a sum image as the base for trimming
    if any(scale > 1):
        newshape = (scale[0] * imshape[0], scale[1] * imshape[1], scale[2] * imshape[2])
        imsum = np.zeros(newshape, dtype=imdtype)
        if not memlim:
            imscl = np.zeros((Nim,) + newshape, dtype=imdtype)
            with trange(Nim, desc="loading-scaling",
                        disable=log.getEffectiveLevel() > logging.INFO,
                        leave=log.getEffectiveLevel() <= logging.INFO) as pbar:
                for i in pbar:
                    imscl[i, :, :, :] = ndi.interpolation.zoom(imin[i, :, :, :], tuple(scale),
                                                               order=int_order)
                    imsum += imscl[i, :, :, :]
        else:
            with trange(Nim, desc="loading-scaling",
                        disable=log.getEffectiveLevel() > logging.INFO,
                        leave=log.getEffectiveLevel() <= logging.INFO) as pbar:
                for i in pbar:
                    if Nim > 50 and using_multiple_files:
                        imin_temp = imio.getnii(imdic['files'][i])
                        imsum += ndi.interpolation.zoom(imin_temp, tuple(scale), order=int_order)
                        log.debug(' image sum: read {}'.format(imdic['files'][i]))
                    else:
                        imsum += ndi.interpolation.zoom(imin[i, :, :, :], tuple(scale),
                                                        order=int_order)
    else:
        imscl = imin
        imsum = np.sum(imin, axis=0)

    # import pdb; pdb.set_trace()
    # ------------------------------------------------------

    if not ref_flag:
        # find the object bounding indexes in x, y and z axes, e.g., ix0-ix1 for the x axis
        qx, qy, qz = im_project3(imsum)

        ix0 = np.argmax(qx > (fmax * np.nanmax(qx)))
        ix1 = ix0 + np.argmin(qx[ix0:] > (fmax * np.nanmax(qx)))

        iy0 = np.argmax(qy > (fmax * np.nanmax(qy)))
        iy1 = iy0 + np.argmin(qy[iy0:] > (fmax * np.nanmax(qy)))

        iz0 = np.argmax(qz > (fmax * np.nanmax(qz)))

        # import pdb; pdb.set_trace()

        # find the maximum voxel range for x and y axes
        IX = ix1 - ix0 + 1
        IY = iy1 - iy0 + 1
        tmp = max(IX, IY)
        # > get the range such that it is divisible by
        # > divdim (64 by default) for GPU execution
        IXY = divdim * ((tmp+divdim-1) // divdim)
        div = (IXY-IX) // 2
        # x
        ix0 -= div
        ix1 += (IXY-IX) - div
        # y
        div = (IXY-IY) // 2
        iy0 -= div
        iy1 += (IXY-IY) - div
        # z
        tmp = (len(qz) - iz0 + 1)
        IZ = divdim * ((tmp+divdim-1) // divdim)
        iz0 -= IZ - tmp + 1

    # save the trimming parameters in a dic
    trimpar = {'x': (ix0, ix1), 'y': (iy0, iy1), 'z': (iz0), 'fmax': fmax, 'scale': scale}

    # new dims (z,y,x)
    newdims = (imsum.shape[0] - iz0, iy1 - iy0 + 1, ix1 - ix0 + 1)
    imtrim = np.zeros((Nim,) + newdims, dtype=imdtype)
    imsumt = np.zeros(newdims, dtype=imdtype)
    # in case of needed padding (negative indx resulting above)
    # the absolute values are supposed to work like padding in case the indx are negative
    iz0s, iy0s, ix0s = iz0, iy0, ix0
    iz0t, iy0t, ix0t = 0, 0, 0
    if iz0 < 0:
        iz0s = 0
        iz0t = abs(iz0)
        log.warning(
            dedent('''\
            -----------------------------------------------------------------
            Correcting for trimming outside the original image (z-axis)'
            -----------------------------------------------------------------'''))

    if iy0 < 0:
        iy0s = 0
        iy0t = abs(iy0)
        log.warning(
            dedent('''\
            -----------------------------------------------------------------
            Correcting for trimming outside the original image (y-axis)'
            -----------------------------------------------------------------'''))

    if ix0 < 0:
        ix0s = 0
        ix0t = abs(ix0)
        log.warning(
            dedent('''\
            -----------------------------------------------------------------
            Correcting for trimming outside the original image (x-axis)
            -----------------------------------------------------------------'''))

    # > in case the upper index goes beyond the scaled but untrimmed image
    iy1t = imsumt.shape[1]
    if iy1 >= imsum.shape[1]:
        iy1t -= iy1 + 1

    # > the same for x
    ix1t = imsumt.shape[2]
    if ix1 >= imsum.shape[2]:
        ix1t -= ix1 + 1

    # first trim the sum image
    imsumt[iz0t:, iy0t:iy1t, ix0t:ix1t] = imsum[iz0s:, iy0s:iy1 + 1, ix0s:ix1 + 1]

    # > new affine matrix for the upscaled and trimmed image
    # > use the scale factor reversely for consistency
    A = np.diag(np.append(sf[::-1], 1.) * np.diag(affine))

    # > note half of new voxel offset is used for the new centre of voxels
    A[0, 3] = affine[0, 3] + A[0, 0] * (ix0-0.5)
    A[1, 3] = affine[1, 3] + (affine[1, 1] * (imshape[1] - 1) - A[1, 1] * (iy1-0.5))
    A[2, 3] = affine[2, 3] - A[1, 1] * 0.5
    A[3, 3] = 1

    # output dictionary
    dctout = {'affine': A, 'trimpar': trimpar, 'imsum': imsumt}

    # NIfTI image description (to be stored in the header)
    niidescr = 'trim(x,y,z):' \
               + str(trimpar['x'])+',' \
               + str(trimpar['y'])+',' \
               + str((trimpar['z'],)) \
               + ';scale='+str(scale) \
               + ';fmx='+str(fmax)

    # > remove brackets and spaces from the file name
    scale_fnm = str(scale).replace('[', '').replace(']', '').replace(' ', '-')

    # store the sum image
    if store_avg and Nim > 1:
        fsum = os.path.join(
            petudir,
            fcomment_pfx + 'avg_trimmed-upsampled-scale-' + scale_fnm + fcomment + '.nii.gz')
        imio.array2nii(imsumt[::-1, ::-1, :], A, fsum, descrip=niidescr)
        log.debug('saved averaged image to: {}'.format(fsum))
        dctout['fsum'] = fsum

    # list of file names for the upsampled and trimmed images
    fpetu = []
    # > perform the trimming and save the intermediate images if requested
    with trange(Nim, desc="finalising trimming/scaling",
                disable=log.getEffectiveLevel() > logging.INFO,
                leave=log.getEffectiveLevel() <= logging.INFO) as pbar:

        for i in pbar:

            # memory saving option, second time doing interpolation
            if memlim:
                if Nim > 50 and using_multiple_files:
                    imin_temp = imio.getnii(imdic['files'][i])
                    im = ndi.interpolation.zoom(imin_temp, tuple(scale), order=int_order)
                    log.debug('image scaling: {}'.format(imdic['files'][i]))
                else:
                    im = ndi.interpolation.zoom(imin[i, :, :, :], tuple(scale), order=int_order)
            else:
                im = imscl[i, :, :, :]

            # trim the scaled image
            imtrim[i, iz0t:, iy0t:iy1t, ix0t:ix1t] = im[iz0s:, iy0s:iy1 + 1, ix0s:ix1 + 1]

            # save the up-sampled and trimmed PET images
            if store_img_intrmd:
                _frm = '_trmfrm' + str(i)
                _fstr = '_trimmed-upsampled-scale-' + scale_fnm + _frm * (Nim > 1) + fcomment
                fpetu.append(os.path.join(petudir, fnms[i] + _fstr + '_i.nii.gz'))
                imio.array2nii(imtrim[i, ::-1, ::-1, :], A, fpetu[i], descrip=niidescr)
                log.debug('saved upsampled PET image to: {}'.format(fpetu[i]))

    if store_img:
        _nfrm = '_nfrm' + str(Nim)
        fim = os.path.join(petudir, fcomment_pfx + 'trimmed-upsampled-scale-' +
                           scale_fnm) + _nfrm * (Nim > 1) + fcomment + '.nii.gz'
        log.info('storing image to:\n{}'.format(fim))

        imio.array2nii(np.squeeze(imtrim[:, ::-1, ::-1, :]), A, fim, descrip=niidescr)
        dctout['fim'] = fim

    # file names (with paths) for the intermediate PET images
    dctout['fimi'] = fpetu
    dctout['im'] = np.squeeze(imtrim)
    dctout['N'] = Nim
    dctout['affine'] = A

    return dctout


# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>

# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>


# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
def iyang(imgIn, krnl, imgSeg, Cnt, itr=5):
    '''
    Partial volume correction using iterative Yang method.
    Arguments:
        imgIn: input image which is blurred due to the PSF of the scanner
        krnl: shift invariant kernel of the PSF
        imgSeg: segmentation into regions starting with 0 (e.g., background)
          and then next integer numbers
        itr: number of iteration (default 5)
    '''
    dim = imgIn.shape
    m = np.int32(np.max(imgSeg))
    m_a = np.zeros((m + 1, itr), dtype=np.float32)

    for jr in range(0, m + 1):
        m_a[jr, 0] = np.mean(imgIn[imgSeg == jr])

    # init output image
    imgOut = np.copy(imgIn)

    # iterative Yang algorithm:
    for i in range(0, itr):
        log.debug('PVC Yang iteration = {}'.format(i))
        # piece-wise constant image
        imgPWC = imgOut
        imgPWC[imgPWC < 0] = 0
        for jr in range(0, m + 1):
            imgPWC[imgSeg == jr] = np.mean(imgPWC[imgSeg == jr])

        # blur the piece-wise constant image
        imgSmo = conv_separable(imgPWC, krnl, dev_id=Cnt['DEVID'])

        # correction factors
        imgCrr = np.ones(dim, dtype=np.float32)
        imgCrr[imgSmo > 0] = imgPWC[imgSmo > 0] / imgSmo[imgSmo > 0]
        imgOut = imgIn * imgCrr
        for jr in range(0, m + 1):
            m_a[jr, i] = np.mean(imgOut[imgSeg == jr])

    return imgOut, m_a


# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>


# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
# G E T   P A R C E L L A T I O N S   F O R   P V C   A N D   R O I   E X T R A C T I O N
# ------------------------------------------------------------------------------------------------------
def pvc_iyang(
    petin,
    mrin,
    Cnt,
    pvcroi,
    krnl,
    itr=5,
    tool='niftyreg',
    faff=None,
    outpath='',
    fcomment='',
    store_img=False,
    store_rois=False,
    matlab_eng_name='',
):
    ''' Perform partial volume (PVC) correction of PET data (petin) using MRI data (mrin).
        The PVC method uses iterative Yang method.
        GPU based convolution is the key routine of the PVC.
        Input:
        -------
        petin:  either a dictionary containing image data, file name and affine transform,
                or a string of the path to the NIfTI file of the PET data.
        mrin: a dictionary of MRI data, including the T1w image, which can be given
                in DICOM (field 'T1DCM') or NIfTI (field 'T1nii').  The T1w image data
                is needed for co-registration to PET if affine is not given in the text
                file with its path in faff.
        Cnt:    a dictionary of paths for third-party tools:
                * dcm2niix: Cnt['DCM2NIIX']
                * niftyreg, resample: Cnt['RESPATH']
                * niftyreg, rigid-reg: Cnt['REGPATH']
        pvcroi: list of regions (also a list) with number label to distinguish
                the parcellations.  The numbers correspond to the image values
                of the parcellated T1w image.  E.g.:
                pvcroi = [
                    [36], # ROI 1 (single parcellation region)
                    [35], # ROI 2
                    [39, 40, 72, 73, 74], # ROI 3 (multiple parcellation regions)
                    ...
                ]
        kernel: the point spread function (PSF) specific for the camera and the object.
                It is given as a 3x17 matrix, a 17-element kernel for each dimension (x,y,z).
                It is used in the GPU-based convolution using separable kernels.
        outpath:path to the output of the resulting PVC images
        faff:   a text file of the affine transformations needed to get the MRI into PET space.
                If not provided, it will be obtained from the performed rigid transformation.
                For this the MR T1w image is required.  If faff and T1w image are not provided,
                it will results in exception/error.
        fcomment:a string used in naming the produced files, helpful for distinguishing them.
        tool:   co-registration tool.  By default it is NiftyReg, but SPM is also
                possible (needs Matlab engine and more validation)
        itr:    number of iterations used by the PVC.  5-10 should be enough (5 default)
    '''
    # get all the input image properties
    if isinstance(petin, dict):
        im = petin['im']
        fpet = petin['fpet']
        B = petin['affine']
    elif isinstance(petin, str) and os.path.isfile(petin):
        imdct = imio.getnii(petin, output='all')
        im = imdct['im']
        B = imdct['affine']
        fpet = petin
    else:
        raise IOError('e> unrecognised input PET file')

    if im.ndim != 3:
        raise IndexError(
            'Only 3D images are expected in this method of partial volume correction.')

    # > avoid registration if the provided parcellation is already in PET space
    # > it is assumed so if the parcellation is given as a file path and no affine is given
    noreg = False
    if isinstance(mrin, str) and os.path.isfile(mrin):
        prcl_dir = os.path.dirname(mrin)
        tmpdct = imio.getnii(mrin, output='all')
        if faff is None and tmpdct['shape'] == imdct['shape']:
            fprcu = mrin
            fprc = mrin
            noreg = True
        elif faff is not None:
            fprc = mrin

    elif isinstance(mrin, dict) and os.path.isfile(mrin['T1lbl']):
        fprc = mrin['T1lbl']
        prcl_dir = os.path.dirname(fprc)

    else:
        raise NameError('e> missing or incorrect labels/parcellations')

    # establish the output folder
    if outpath == '':
        oprcl = os.path.join(prcl_dir, 'PVC-preprocessed')
        opvc = os.path.join(os.path.dirname(fpet), 'PVC')
    else:
        oprcl = os.path.join(outpath, 'PVC-preprocessed')
        opvc = os.path.join(outpath, 'PVC')

    # > create folders
    imio.create_dir(oprcl)
    imio.create_dir(opvc)
    if store_rois:
        orois = os.path.join(opvc, 'ROIs')
        imio.create_dir(orois)

    # > output dictionary
    outdct = {}

    # =================================================================
    # > if affine transformation (faff) is not given then register T1 to PET
    # and resample parcellations
    if not noreg and faff is None:
        ft1w = imio.pick_t1w(mrin)
        if tool == 'spm':
            regdct = regseg.coreg_spm(fpet, ft1w, matlab_eng_name=matlab_eng_name,
                                      fcomment=fcomment,
                                      outpath=os.path.join(outpath, 'PET', 'positioning'))
        elif tool == 'niftyreg':
            regdct = regseg.affine_niftyreg(
                fpet,
                ft1w,
                outpath=os.path.join(outpath, 'PET', 'positioning'),
                fcomment=fcomment,
                executable=Cnt['REGPATH'],
                omp=multiprocessing.cpu_count() / 2,
                rigOnly=True,
                affDirect=False,
                maxit=5,
                speed=True,
                pi=50,
                pv=50,
                smof=0,
                smor=0,
                rmsk=True,
                fmsk=True,
                rfwhm=15.,                                           # millilitres
                rthrsh=0.05,
                ffwhm=15.,                                           # millilitres
                fthrsh=0.05)
        faff = regdct['faff']

    # resample the T1/labels to upsampled PET
    # file name of the parcellation (e.g., GIF-based) upsampled to PET
    if faff is not None and os.path.isfile(faff):
        fprcu = os.path.join(
            oprcl,
            os.path.basename(fprc.split('.')[0] + '_registered_trimmed' + fcomment + '.nii.gz'))

        if tool == 'niftyreg':
            if os.path.isfile(Cnt['RESPATH']):
                cmd = [
                    Cnt['RESPATH'], '-ref', fpet, '-flo', fprc, '-trans', faff, '-res', fprcu,
                    '-inter', '0']
                if log.getEffectiveLevel() >= logging.INFO:
                    cmd.append('-voff')
                run(cmd)
            else:
                raise IOError('e> path to resampling executable is incorrect!')
        elif tool == 'spm':
            regseg.resample_spm(
                fpet,
                fprc,
                faff,
                fimout=fprcu,
                matlab_eng_name=matlab_eng_name,
                intrp=0.,
                del_ref_uncmpr=True,
                del_flo_uncmpr=True,
                del_out_uncmpr=True,
            )
    # =================================================================

    # > get the parcellation labels in the upsampled PET space
    prcdct = imio.getnii(fprcu, output='all')
    prcu = prcdct['im']

    # > path to parcellations in NIfTI format
    prcl_pth = os.path.split(fprc)

    # --------------------------------------------------------------------------
    # > get the parcellation specific for PVC based on the current parcellations
    imgroi = prcu.copy()
    imgroi[:] = 0

    # > number of segments, without the background
    nSeg = len(pvcroi)

    # > create the image of numbered parcellations
    for k in range(nSeg):
        for m in pvcroi[k]:
            imgroi[prcu == m] = k + 1

    # > save the PCV ROIs to a new NIfTI file
    if store_rois:
        froi = os.path.join(orois, prcl_pth[1].split('.nii')[0] + '_PVC-ROIs-inPET.nii.gz')
        imio.array2nii(
            imgroi, prcdct['affine'], froi,
            trnsp=(prcdct['transpose'].index(0), prcdct['transpose'].index(1),
                   prcdct['transpose'].index(2)), flip=prcdct['flip'])
        outdct['froi'] = froi
    # --------------------------------------------------------------------------

    # --------------------------------------------------------------------------
    # run iterative Yang PVC
    imgpvc, m_a = iyang(im, krnl, imgroi, Cnt, itr=itr)
    # --------------------------------------------------------------------------

    outdct['im'] = imgpvc
    outdct['imroi'] = imgroi
    outdct['fprc'] = fprcu
    outdct['imprc'] = prcu
    if not noreg:
        outdct['faff'] = faff

    if store_img:
        fpvc = os.path.join(
            opvc,
            os.path.split(fpet)[1].split('.nii')[0] + '_PVC' + fcomment + '.nii.gz')
        imio.array2nii(imgpvc[::-1, ::-1, :], B, fpvc, descrip='pvc=iY')
        outdct['fpet'] = fpvc

    return outdct


# =============================================================
# Convert CT units (HU) to PET mu-values
def ct2mu(im):
    '''HU units to 511keV PET mu-values
        https://link.springer.com/content/pdf/10.1007%2Fs00259-002-0796-3.pdf
        C. Burger, et al., PET attenuation coefficients from CT images,
    '''

    # convert nans to -1024 for the HU values only
    im[np.isnan(im)] = -1024
    # constants
    muwater = 0.096
    mubone = 0.172
    rhowater = 0.184
    rhobone = 0.428
    uim = np.zeros(im.shape, dtype=np.float32)
    uim[im <= 0] = muwater * (1 + im[im <= 0] * 1e-3)
    uim[im > 0] = muwater + im[im > 0] * (rhowater * (mubone-muwater) / (1e3 * (rhobone-rhowater)))
    # remove negative values
    uim[uim < 0] = 0
    return uim


# =============================================================


# ==============================================================================
def centre_mass_img(imdct, output='mm'):
    ''' Calculate the centre of mass of an image along each axes (x,y,z),
        separately.
        Arguments:
        imdct - the image dictionary with the image and header data.
        Output the list of the centre of mass for each axis.
    '''

    # > initialise centre of mass array in mm and in voxel indexes
    com = np.zeros(3, dtype=np.float32)
    icom = np.zeros(3, dtype=np.float32)
    # > total image sum
    imsum = np.sum(imdct['im'])

    for ind_ax in [-1, -2, -3]:
        # > list of axes
        axs = list(range(imdct['im'].ndim))
        del axs[ind_ax]
        # > indexed centre of mass
        icom[ind_ax] = np.sum(
            np.sum(imdct['im'], axis=tuple(axs)) * np.arange(imdct['shape'][ind_ax])) / imsum
        # > centre of mass in mm (zyx)
        com[ind_ax] = icom[ind_ax] * imdct['voxsize'][ind_ax]

    # > correct due to flipped indexing and world mm
    com[-1] = imdct['shape'][-1] * imdct['voxsize'][-1] - com[-1]
    com[-2] = imdct['shape'][-2] * imdct['voxsize'][-2] - com[-2]
    com[-3] = imdct['shape'][-3] * imdct['voxsize'][-3] - com[-3]

    if output == 'mm':
        return com
    elif output == 'vox':
        return icom
    else:
        raise ValueError('unrecognised output option')


# ==============================================================================


# ==============================================================================
def nii_modify(nii, fimout='', outpath='', fcomment='', voxel_range=None):
    '''
    Modify the NIfTI image given either as a file path or a dictionary,
    obtained by nimpa.getnii(file_path).
    '''
    if voxel_range is None:
        voxel_range = []
    if isinstance(nii, str) and os.path.isfile(nii):
        dctnii = imio.getnii(nii, output='all')
        fnii = nii
    if isinstance(nii, dict) and 'im' in nii:
        dctnii = nii
        if 'fim' in dctnii:
            fnii = dctnii['fim']

    # --------------------------------------------------------------------------
    # > output path
    if outpath == '' and fimout != '' and '/' in fimout:
        opth = os.path.dirname(fimout)
        if opth == '' and isinstance(fnii, str) and os.path.isfile(fnii):
            opth = os.path.dirname(nii)
        fimout = os.path.basename(fimout)

    elif outpath == '' and isinstance(fnii, str) and os.path.isfile(fnii):
        opth = os.path.dirname(fnii)
    else:
        opth = outpath
    imio.create_dir(opth)

    # > output floating and affine file names
    if fimout == '':

        if fcomment == '':
            fcomment += '_nimpa-modified'

        fout = os.path.join(opth, os.path.basename(fnii).split('.nii')[0] + fcomment + '.nii.gz')
    else:
        fout = os.path.join(opth, fimout.split('.')[0] + '.nii.gz')
    # --------------------------------------------------------------------------

    # > reduce the max value to 255
    if voxel_range and len(voxel_range) == 1:
        im = voxel_range[0] * dctnii['im'] / np.max(dctnii['im'])
    elif voxel_range and len(voxel_range) == 2:
        # > normalise into range 0-1
        im = (dctnii['im'] - np.min(dctnii['im'])) / np.ptp(dctnii['im'])
        # > convert to voxel_range
        im = voxel_range[0] + im * (voxel_range[1] - voxel_range[0])
    else:
        return None

    # > output file name for the extra reference image
    imio.array2nii(
        im, dctnii['affine'], fout,
        trnsp=(dctnii['transpose'].index(0), dctnii['transpose'].index(1),
               dctnii['transpose'].index(2)), flip=dctnii['flip'])

    return {'fim': fout, 'im': im, 'affine': dctnii['affine']}


# ==============================================================================


# > used for cutting out part of image for face de-identification
def im_cut(im, i_cut, fout=None):
    ''' cut the part of image like the face for de-identification purposes.
        assumes that image is in the form of im[z,y,x].
    '''

    if fout is None:
        save_nii = False
    else:
        save_nii = True

    # > output dictionary
    out = {}

    if isinstance(im, str):
        imdct = imio.getnii(im, output='all')
        img = imdct['im']
        save_nii = True
        fout_s = os.path.split(im)
        fout = os.path.join(fout_s[0], fout_s[1].split('.nii')[0] + '_cut.nii.gz')
        out['fim'] = fout
    elif isinstance(im, (np.ndarray, np.generic)):
        img = im
    else:
        raise ValueError('unrecognised image input')

    # prj = im_project3(img)
    # plot(prj[0]); plot(prj[1]); plot(prj[2])

    img[:, :i_cut, :] = 0

    if save_nii:
        imio.array2nii(
            img, imdct['affine'], fout,
            trnsp=(imdct['transpose'].index(0), imdct['transpose'].index(1),
                   imdct['transpose'].index(2)), flip=imdct['flip'])
        out['fim'] = fout

    out['im'] = img

    return out


# ==============================================================================

#  ____________________________________________________________________________
# |                                                                            |
# |                 M R   B I A S   C O R R E C T I O N                        |
# |____________________________________________________________________________|


def bias_field_correction(fmr, fimout='', outpath='', fcomment='_N4bias', executable='',
                          exe_options=None, sitk_image_mask=True, verbose=False, Cnt=None):
    ''' Correct for bias field in MR image(s) given in <fmr> as a string
        (single file) or as a list of strings (multiple files).

        Output dictionary with the bias corrected file names.

        Options:
        - fimout:       The name (with path) of the output file.  It's
                        ignored when multiple files are given as input.  If
                        given for a single file name, the <outpath> and
                        <fcomment> options are ignored.
        - outpath:      Path to the output folder
        - fcomment:     A prefix comment to the file name
        - executable:   The path to the executable;  if 'sitk' is given instead
                        of the path, the Python module SimpleITK will be
                        used if it is available.
        - exe_options:  Options for the executable in the form of a list of
                        strings.
        - sitk_image_mask:  Image masking will be used if SimpleITK is
                            chosen.
    '''
    if exe_options is None:
        exe_options = []
    if Cnt is None:
        Cnt = {}

    if executable == 'sitk' and 'SimpleITK' not in sys.modules and not sitk_flag:
        raise ImportError(
            dedent('''\
            If SimpleITK module is required for bias correction, it needs to be
            first installed using this command:
            conda install -c simpleitk simpleitk
            or pip install SimpleITK'''))

    # --------------------------------------------------------------------------
    # INPUT
    # --------------------------------------------------------------------------
    # > path to a single file
    if isinstance(fmr, str) and os.path.isfile(fmr):
        fins = [fmr]

    # > list of file paths
    elif isinstance(fmr, list) and all(map(os.path.isfile, fmr)):
        fins = fmr
        log.info('multiple input files => ignoring the single output file name.')
        fimout = ''

    # > path to a folder
    elif isinstance(fmr, str) and os.path.isdir(fmr):
        fins = [os.path.join(fmr, f) for f in os.listdir(fmr) if f.endswith(('.nii', '.nii.gz'))]
        log.info('multiple input files from input folder.')
        fimout = ''

    else:
        raise ValueError('could not decode the input of floating images.')
    # --------------------------------------------------------------------------

    # --------------------------------------------------------------------------
    # OUTPUT
    # --------------------------------------------------------------------------
    # > output path
    if outpath == '' and fimout != '':
        opth = os.path.dirname(fimout)
        if opth == '':
            opth = os.path.dirname(fmr)
            fimout = os.path.join(opth, fimout)
        n4opth = opth
        fcomment = ''
    elif outpath == '':
        opth = os.path.dirname(fmr)
        # > N4 bias correction specific folder
        n4opth = os.path.join(opth, 'N4bias')
    else:
        opth = outpath
        # > N4 bias correction specific folder
        n4opth = os.path.join(opth, 'N4bias')

    imio.create_dir(n4opth)

    outdct = {}
    # --------------------------------------------------------------------------

    for fin in fins:
        log.debug('input for bias correction:\n{}'.format(fin))

        if fimout == '':
            # split path
            fspl = os.path.split(fin)

            # N4 bias correction file output paths
            fn4 = os.path.join(n4opth, fspl[1].split('.nii')[0] + fcomment + '.nii.gz')
        else:
            fn4 = fimout

        if not os.path.exists(fn4):

            if executable == 'sitk':
                # =============================================
                # SimpleITK Bias field correction for T1 and T2
                # =============================================
                # > initialise the corrector
                corrector = sitk.N4BiasFieldCorrectionImageFilter()
                # numberFilltingLevels = 4

                # read input file
                im = sitk.ReadImage(fin)

                # > create a object specific mask
                fmsk = os.path.join(n4opth, fspl[1].split('.nii')[0] + '_sitk_mask.nii.gz')
                msk = sitk.OtsuThreshold(im, 0, 1, 200)
                sitk.WriteImage(msk, fmsk)

                # > cast to 32-bit float
                im = sitk.Cast(im, sitk.sitkFloat32)

                # ------------------------------------------
                log.info('correcting bias field for {}'.format(fin))
                n4out = corrector.Execute(im, msk)
                sitk.WriteImage(n4out, fn4)
                # ------------------------------------------

                if sitk_image_mask:
                    outdct.setdefault('fmsk', [])
                    outdct['fmsk'].append(fmsk)

            elif os.path.basename(executable)=='N4BiasFieldCorrection' \
                    and os.path.isfile(executable):

                cmd = [executable, '-i', fin, '-o', fn4]

                if verbose and os.path.basename(executable) == 'N4BiasFieldCorrection':
                    cmd.extend(['-v', '1'])

                cmd.extend(exe_options)

                run(cmd)

                if 'command' not in outdct:
                    outdct['command'] = []
                outdct['command'].append(cmd)

            elif os.path.isfile(executable):

                cmd = [executable]
                cmd.extend(exe_options)
                run(cmd)

                if 'command' not in outdct:
                    outdct['command'] = cmd

        else:
            log.info('N4 bias corrected file seems already existing.')

        # > output to dictionary
        outdct.setdefault('fim', [])
        outdct['fim'].append(fn4)

    return outdct


# ------------------------------------------------------------------------------------
def pet2pet_rigid(fref, fflo, Cnt, outpath='', rmsk=True, rfwhm=15., rthrsh=0.05, pi=50, pv=50,
                  smof=0, smor=0):

    # create a folder for PET images registered to ref PET
    if outpath == '':
        outpath = os.path.dirname(fflo)
    odir = os.path.join(outpath, 'PET2PET')
    imio.create_dir(odir)

    if rmsk:
        fimdir = os.path.join(odir, 'tmp')
        imio.create_dir(fimdir)
        fmsk = os.path.join(fimdir, 'rmask.nii.gz')
        imdct = imio.getnii(fref, output='all')
        smoim = ndi.filters.gaussian_filter(imdct['im'],
                                            imio.fwhm2sig(rfwhm, voxsize=imdct['affine'][0, 0]),
                                            mode='mirror')
        thrsh = rthrsh * smoim.max()
        immsk = np.int8(smoim > thrsh)
        for iz in range(immsk.shape[0]):
            for iy in range(immsk.shape[1]):
                ix0 = np.argmax(immsk[iz, iy, :] > 0)
                ix1 = immsk.shape[2] - np.argmax(immsk[iz, iy, ::-1] > 0)
                if (ix1 - ix0) > immsk.shape[2] - 10: continue
                immsk[iz, iy, ix0:ix1] = 1
        imio.array2nii(immsk[::-1, ::-1, :], imio.getnii_affine(fref), fmsk)

    # output in register with ref PET
    fout = os.path.join(odir, 'PET-r-to-' + os.path.basename(fref).split('.')[0] + '.nii.gz')
    # text file for the affine transform T1w->PET
    faff = os.path.join(odir, 'affine-PET-r-to-' + os.path.basename(fref).split('.')[0] + '.txt')
    # call the registration routine
    if os.path.isfile(Cnt['REGPATH']):
        cmd = [
            Cnt['REGPATH'], '-ref', fref, '-flo', fflo, '-rigOnly', '-speeeeed', '-aff', faff,
            '-pi',
            str(pi), '-pv',
            str(pv), '-smooF',
            str(smof), '-smooR',
            str(smor), '-res', fout]
        if rmsk:
            cmd.append('-rmask')
            cmd.append(fmsk)
        if log.getEffectiveLevel() >= logging.INFO:
            cmd.append('-voff')
        log.info('Executing command:\n{}'.format(cmd))
        run(cmd)
    else:
        log.error('path to registration executable is incorrect!')
        raise Exception('No registration executable found')

    return faff, fout


# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>


def mr2pet_rigid(fpet, mridct, Cnt, outpath='', fcomment='', rmsk=True, rfwhm=15., rthrsh=0.05,
                 fmsk=True, ffwhm=15., fthrsh=0.05, pi=50, pv=50, smof=0, smor=0):
    # create output path if given
    if outpath != '':
        imio.create_dir(outpath)

    # --- MR T1w
    if 'T1nii' in mridct and os.path.isfile(mridct['T1nii']):
        ft1w = mridct['T1nii']
    elif 'T1bc' in mridct and os.path.isfile(mridct['T1bc']):
        ft1w = mridct['T1bc']
    elif 'T1DCM' in mridct and os.path.exists(mridct['MRT1W']):
        # create file name for the converted NIfTI image
        fnii = 'converted'
        run([Cnt['DCM2NIIX'], '-f', fnii, mridct['T1nii']])
        ft1nii = glob(os.path.join(mridct['T1nii'], '*converted*.nii*'))
        ft1w = ft1nii[0]
    else:
        raise ValueError('disaster: no T1w image!')

    # create a folder for MR images registered to PET
    if outpath != '':
        mrodir = os.path.join(outpath, 'T1w2PET')
        fimdir = os.path.join(outpath, os.path.join('T1w2PET', 'tmp'))
    else:
        mrodir = os.path.join(os.path.dirname(ft1w), 'mr2pet')
        fimdir = os.path.join(os.path.basename(ft1w), 'tmp')
    imio.create_dir(mrodir)
    imio.create_dir(fimdir)

    if rmsk:
        f_rmsk = os.path.join(fimdir, 'rmask.nii.gz')
        imdct = imio.getnii(fpet, output='all')
        smoim = ndi.filters.gaussian_filter(
            imdct['im'], imio.fwhm2sig(rfwhm, voxsize=abs(imdct['affine'][0, 0])), mode='mirror')
        thrsh = rthrsh * smoim.max()
        immsk = np.int8(smoim > thrsh)
        immsk = regseg.imfill(immsk)
        imio.array2nii(immsk[::-1, ::-1, :], imdct['affine'], f_rmsk)
    if fmsk:
        f_fmsk = os.path.join(fimdir, 'fmask.nii.gz')
        imdct = imio.getnii(ft1w, output='all')
        smoim = ndi.filters.gaussian_filter(
            imdct['im'], imio.fwhm2sig(ffwhm, voxsize=abs(imdct['affine'][0, 0])), mode='mirror')
        thrsh = fthrsh * smoim.max()
        immsk = np.int8(smoim > thrsh)
        immsk = regseg.imfill(immsk)
        imio.array2nii(immsk[::-1, ::-1, :], imdct['affine'], f_fmsk)

    # if provided, separate the comment with underscore
    if fcomment != '': fcomment = '_' + fcomment
    # output for the T1w in register with PET
    ft1out = os.path.join(
        mrodir, 'T1w-r-to-' + os.path.basename(fpet).split('.')[0] + fcomment + '.nii.gz')
    # text file for the affine transform T1w->PET
    faff = os.path.join(
        mrodir, 'affine-T1w-r-to-' + os.path.basename(fpet).split('.')[0] + fcomment + '.txt')
    # call the registration routine
    if os.path.isfile(Cnt['REGPATH']):
        cmd = [
            Cnt['REGPATH'], '-ref', fpet, '-flo', ft1w, '-rigOnly', '-speeeeed', '-aff', faff,
            '-pi',
            str(pi), '-pv',
            str(pv), '-smooF',
            str(smof), '-smooR',
            str(smor), '-res', ft1out]
        if rmsk:
            cmd.append('-rmask')
            cmd.append(f_rmsk)
        if fmsk:
            cmd.append('-fmask')
            cmd.append(f_fmsk)
        if not Cnt.get('VERBOSE', True):
            cmd.append('-voff')
        log.info('Executing command:\n{}'.format(cmd))
        run(cmd)
    else:
        raise IOError('path to registration executable is incorrect!')

    return faff



#!/usr/bin/env python
"""initialise the NiftyPET NIMPA package"""
__author__ = "Pawel J. Markiewicz", "Casper O. da Costa-Luis"
__copyright__ = "Copyright 2021"
# version detector. Precedence: installed dist, git, 'UNKNOWN'
try:
    from ._dist_ver import __version__
except ImportError:
    try:
        from setuptools_scm import get_version

        __version__ = get_version(root="../..", relative_to=__file__)
    except (ImportError, LookupError):
        __version__ = "UNKNOWN"
__all__ = [
    # gpu utils
    'cs', 'dev_info', 'gpuinfo',
    # utils
    'LOG_FORMAT', 'LogHandler',
    # config
    'path_resources', 'resources', 'cmake_prefix',
    # core
    'create_disk', 'imdiff', 'imscroll', 'profile_points', 'imtrimup',
    'affine_fsl', 'affine_niftyreg', 'array2nii', 'bias_field_correction',
    'centre_mass_img', 'coreg_spm', 'coreg_vinci', 'create_dir', 'create_mask',
    'ct2mu', 'dcm2im', 'dcm2nii', 'dcmanonym', 'dcminfo', 'dcmsort',
    'dice_coeff', 'dice_coeff_multiclass', 'fwhm2sig', 'getnii', 'getnii_descr',
    'im_cut', 'imfill', 'imsmooth', 'iyang', 'motion_reg', 'nii_gzip',
    'nii_modify', 'nii_ugzip', 'niisort', 'orientnii', 'pet2pet_rigid',
    'pick_t1w', 'psf_gaussian', 'psf_measured', 'pvc_iyang', 'realign_mltp_spm',
    'resample_fsl', 'resample_mltp_spm', 'resample_niftyreg', 'resample_spm',
    'resample_vinci', 'time_stamp'] # yapf: disable

from pkg_resources import resource_filename

from niftypet.ninst import cudasetup as cs
from niftypet.ninst.dinf import dev_info, gpuinfo
from niftypet.ninst.tools import LOG_FORMAT, LogHandler, path_resources, resources

from .img import create_disk, imdiff, imscroll, profile_points
from .prc import imtrimup  # for backward compatibility
from .prc import (
    affine_fsl,
    affine_niftyreg,
    array2nii,
    bias_field_correction,
    centre_mass_img,
    coreg_spm,
    coreg_vinci,
    create_dir,
    create_mask,
    ct2mu,
    dcm2im,
    dcm2nii,
    dcmanonym,
    dcminfo,
    dcmsort,
    dice_coeff,
    dice_coeff_multiclass,
    fwhm2sig,
    getnii,
    getnii_descr,
    im_cut,
    imfill,
    imsmooth,
    iyang,
    motion_reg,
    nii_gzip,
    nii_modify,
    nii_ugzip,
    niisort,
    orientnii,
    pet2pet_rigid,
    pick_t1w,
    psf_gaussian,
    psf_measured,
    pvc_iyang,
    realign_mltp_spm,
    resample_fsl,
    resample_mltp_spm,
    resample_niftyreg,
    resample_spm,
    resample_vinci,
    time_stamp,
)

# for use in `cmake -DCMAKE_PREFIX_PATH=...`
cmake_prefix = resource_filename(__name__, "cmake")



# initialise the module folder
__all__ = [
    # imio
    'array2nii', 'create_dir', 'dcm2im', 'dcm2nii', 'dcmanonym', 'dcminfo',
    'dcmsort', 'fwhm2sig', 'getnii', 'getnii_descr', 'nii_gzip', 'nii_ugzip',
    'niisort', 'orientnii', 'pick_t1w', 'time_stamp',
    # prc
    'bias_field_correction', 'centre_mass_img', 'ct2mu', 'im_cut', 'imsmooth',
    'imtrimup', 'iyang', 'nii_modify', 'pet2pet_rigid', 'psf_gaussian',
    'psf_measured', 'pvc_iyang',
    # regseg
    'affine_fsl', 'affine_niftyreg', 'coreg_spm', 'coreg_vinci', 'create_mask',
    'dice_coeff', 'dice_coeff_multiclass', 'imfill', 'motion_reg',
    'realign_mltp_spm', 'resample_fsl', 'resample_mltp_spm',
    'resample_niftyreg', 'resample_spm', 'resample_vinci'] # yapf: disable

from .imio import (
    array2nii,
    create_dir,
    dcm2im,
    dcm2nii,
    dcmanonym,
    dcminfo,
    dcmsort,
    fwhm2sig,
    getnii,
    getnii_descr,
    nii_gzip,
    nii_ugzip,
    niisort,
    orientnii,
    pick_t1w,
    time_stamp,
)

# will be deprecated
from .prc import (
    bias_field_correction,
    centre_mass_img,
    ct2mu,
    im_cut,
    imsmooth,
    imtrimup,
    iyang,
    nii_modify,
    pet2pet_rigid,
    psf_gaussian,
    psf_measured,
    pvc_iyang,
)
from .regseg import (
    affine_fsl,
    affine_niftyreg,
    coreg_spm,
    coreg_vinci,
    create_mask,
    dice_coeff,
    dice_coeff_multiclass,
    imfill,
    motion_reg,
    realign_mltp_spm,
    resample_fsl,
    resample_mltp_spm,
    resample_niftyreg,
    resample_spm,
    resample_vinci,
)



"""
NIMPA: functions for neuro image processing and analysis
Generates images.
"""
import logging
import math

import numpy as np
import scipy.ndimage as ndi

try:
    from miutil.plot import imscroll
except ImportError as err: # NOQA: F841

    def imscroll(*_, **__):
        """delay matplotlib import error for later"""
        raise err  # NOQA: F821


from ..prc import imio

log = logging.getLogger(__name__)


def absmax(a):
    amax = a.max()
    amin = a.min()
    return np.where(-amin > amax, amin, amax)


def create_disk(shape_in, r=1, a=0, b=0, gen_scale=1, threshold=None):
    if len(shape_in) == 2:
        shape = (1,) + shape_in
    if len(shape_in) == 3:
        shape = shape_in

    imsk = np.zeros((gen_scale * shape[1], gen_scale * shape[2]), dtype=np.float32)
    for t in np.arange(0, math.pi, math.pi / (gen_scale*400)):
        x = gen_scale * r * np.cos(t) + gen_scale*a
        y = gen_scale * r * np.sin(t) + gen_scale*b

        for ys in np.arange(-y + 2*b*gen_scale, y, 0.5):
            v = 0.5 * gen_scale * shape[1] - np.ceil(ys)
            u = 0.5 * gen_scale * shape[2] + np.floor(x)
            imsk[v.astype(np.int16), u.astype(np.int16)] = 1.

    if gen_scale > 1:
        imsk = ndi.interpolation.zoom(imsk, gen_scale**-1, order=1)

    if threshold:
        imsk = imsk > threshold

    if len(shape_in) == 3:
        msk = np.repeat(imsk.reshape((1, shape[1], shape[1])), shape[0], axis=0)
    elif len(shape_in) == 2:
        msk = imsk
    return msk


def profile_points(im, p0, p1, steps=100):
    p = np.array([p1[0] - p0[0], p1[1] - p0[1]])
    nrm = np.sum(p**2)**.5
    p = p / nrm

    tt = np.linspace(0, nrm, steps)
    profile = np.zeros(len(tt), dtype=im.dtype)

    c = 0
    for t in tt:
        u = t * p[0] + p0[0]
        v = t * p[1] + p0[1]

        profile[c] = im[int(round(v)), int(round(u))]
        c += 1

    return profile


# ----------------------------------------------------------------------------------------------------------------------------
def imdiff(imref, imnew, verbose=False, plot=False, cmap='bwr'):
    """
    Compare the new image (imnew) to the reference image (imref),
    returning (and optionallt plotting) the difference.
    """
    if isinstance(imref, str):
        imref = imio.getnii(imref)
        log.info('using NIfTI files as image input for the reference')
    elif isinstance(imref, (np.ndarray, np.generic)):
        log.info('using Numpy arrays as input for the reference image')
    elif isinstance(imref, dict):
        imref = imref['im']
        log.info('using the input dictionary with an image for reference')

    if isinstance(imnew, str):
        imnew = imio.getnii(imnew)
        log.info('using NIfTI files as input for the new image')
    elif isinstance(imnew, (np.ndarray, np.generic)):
        log.info('using Numpy arrays as input for the new image')
    elif isinstance(imnew, dict):
        imnew = imnew['im']
        log.info('using the input dictionary with a new image for comparison')

    # -----------------------------------------------------------------------------------
    # MAXIMUM PROJECTION IMAGE
    if plot:
        import matplotlib.pyplot as plt

        # > threshold percentage used in plotting (helps ignoring singular hot values)
        maxproj_thrshl = 0.7

        # > maximum projection along axis ax
        def maxproj(imdiff, ax):
            # > maximum projection image
            imp = np.max(imdiff, axis=ax)
            # > minimum projection image
            imn = np.min(imdiff, axis=ax)
            # > max mask
            mmsk = imp > abs(imn)
            # > form the highest intensity projection image (positive and negative)
            im = imn.copy()
            im[mmsk] = imp[mmsk]
            # > maximum/minimum value in the difference image for a symmetrical colour map
            valmax = max(np.max(imp), np.min(imn))
            return im, valmax

    # -----------------------------------------------------------------------------------

    if imnew.ndim == 3 and imref.ndim == 3 and imnew.shape == imref.shape:
        imnew = imnew[np.newaxis, ...]
        imref = imref[np.newaxis, ...]
        log.info('using 3D images as input')
        Nim = 1
    elif imnew.ndim == 4 and imref.ndim == 4 and imnew.shape == imref.shape:
        log.info('using 4D images as input')
        # > this assumes that axis 0 encodes image frames (e.g., along time)
        Nim = imnew.shape[0]
    else:
        raise ValueError('the input images have to be of the same dimensions and shape')

    mape = np.zeros(Nim)
    mae = np.zeros(Nim)
    mad = np.zeros(Nim)

    for i in range(Nim):
        # > maximum voxel value of reference image
        mx = np.max(imref[i, ...])

        # > create a image mask based on the max the 0.1% value
        msk = imref[i, ...] > 0.001 * mx

        # > image difference
        imdiff = imref[i, ...] - imnew[i, ...]

        # > mean absolute percentage difference
        mape[i] = np.mean(abs(imdiff[msk] / imref[i, msk])) * 100

        # > mean absolute error
        mae[i] = np.mean(abs(imdiff[msk]))

        # > maximum absolute difference
        mad[i] = np.max(abs(imdiff[msk]))

        if verbose:
            print('---------------------------------------------------------------')
            print('>> frame {}: mean absolute relative image difference [%]:'.format(i))
            print(mape[i])

            print('>> frame {}: mean absolute image difference:'.format(i))
            print(mae[i])

            print('>> frame {}:maximum absolute image difference:'.format(i))
            print(mad[i])

        if plot:

            fig = plt.figure(figsize=(12, 6))
            fig.suptitle('maximum absolute difference projection along 3 axes', fontsize=12,
                         fontweight='bold')

            im, valmax = maxproj(imdiff, 0)

            plt.subplot(131)
            plt.imshow(im, cmap=cmap, vmax=maxproj_thrshl * valmax, vmin=-maxproj_thrshl * valmax)
            plt.colorbar()

            im, valmax = maxproj(imdiff, 1)

            plt.subplot(132)
            plt.imshow(im, cmap=cmap, vmax=maxproj_thrshl * valmax, vmin=-maxproj_thrshl * valmax)
            # plt.colorbar()

            im, valmax = maxproj(imdiff, 2)
            plt.subplot(133)
            plt.imshow(im, cmap=cmap, vmax=maxproj_thrshl * valmax, vmin=-maxproj_thrshl * valmax)
            # plt.colorbar()

            plt.show()

    if Nim > 1:
        out = {'mape': mape, 'mae': mae, 'mad': mad}
    else:
        out = {'mape': mape[0], 'mae': mae[0], 'mad': mad[0]}

    return out



import logging

import numpy as np
from pytest import fixture, importorskip, mark

from niftypet.nimpa.prc import prc

cu = importorskip("cuvec")
improc = importorskip("niftypet.nimpa.prc.improc")


def check_conv(src, dst):
    err = "\nsrc:\n%s\ndst:\n%s" % (src[2, :3, :3], dst[2, :3, :3])
    assert dst[2, 2, 2] == 1 / 8, err
    assert dst[1, 2, 2] == 1 / 16, err
    assert dst[1, 2, 1] == 1 / 32, err
    assert dst[1, 1, 1] == 1 / 64, err


def rmse(x, y):
    return (((x - y)**2).mean() / (y**2).mean())**0.5


@fixture(scope="module")
def knl():
    res = cu.zeros((3, 17), dtype='float32')
    res[:, 17//2 - 1] = 0.25
    res[:, 17 // 2] = 0.5
    res[:, 17//2 + 1] = 0.25
    return res


def test_convolve_unpadded(knl):
    src = cu.zeros((64, 64, 64), dtype='float32')
    src[2, 2, 2] = 1

    # won't memcopy since already cudaMallocManaged
    dst = cu.asarray(improc.convolve(src.cuvec, knl.cuvec, log=logging.DEBUG))
    check_conv(src, dst)


def test_convolve_autopad(knl):
    src_np = np.zeros((27, 13, 7), dtype='float32')
    src_np[2, 2, 2] = 1

    src = cu.asarray(src_np) # memcopy to cudaMallocManaged
    assert knl.dtype == src.dtype

    # won't memcopy since already cudaMallocManaged
    dst = cu.asarray(improc.convolve(src.cuvec, knl.cuvec, log=logging.DEBUG))
    check_conv(src, dst)

    # in-place
    dst[:] = 0
    assert not dst.any()
    improc.convolve(src.cuvec, knl.cuvec, log=logging.DEBUG, output=dst.cuvec, memset=False)
    check_conv(src, dst)

    # in-place with auto memset
    dst[2, 2, 2] = 0
    assert dst[2, 2, 2] == 0
    improc.convolve(src.cuvec, knl.cuvec, log=logging.DEBUG, output=dst.cuvec)
    check_conv(src, dst)


@mark.parametrize("knl_size", [(3, 17), (2, 17), (3, 5), (2, 5), (1, 3)])
def test_conv_separable(knl_size):
    knl = np.random.random(knl_size)
    src = np.random.random((64,) * knl_size[0]).astype('float32')
    dst_gpu = prc.conv_separable(src, knl)
    dst_cpu = prc.conv_separable(src, knl, dev_id=False)
    assert hasattr(dst_gpu, 'cuvec') or knl_size[0] != 3
    assert not hasattr(dst_cpu, 'cuvec')
    assert rmse(dst_gpu, dst_cpu) < 1e-7


@mark.parametrize("half_width", [0, 1, 2])
@mark.parametrize("sigma", [0.5, 1, 2])
@mark.parametrize("width", [10, 125])
def test_nlm(half_width, sigma, width):
    src = np.random.random((width,) * 3).astype('float32')
    ref = np.random.random((width,) * 3)
    dst_gpu = prc.nlm(src, ref, half_width=half_width, sigma=sigma)
    assert hasattr(dst_gpu, 'cuvec')
    assert (dst_gpu - src).mean() < 1e-2


if __name__ == "__main__":
    from sys import version_info
    from textwrap import dedent

    from argopt import argopt
    from tqdm import trange
    logging.basicConfig(level=logging.WARNING)

    parser = argopt(
        dedent("""\
        Usage:
            test_improc [options]

        Options:
            -r REP, --repeats REP  : [default: 10:int]
        """))
    if version_info[:2] >= (3, 7):
        subs = parser.add_subparsers(required=True)
    else:
        subs = parser.add_subparsers()

    def sub_parser(prog=None, **kwargs):
        return subs.add_parser(prog, **kwargs)

    def conv(args):
        """\
        Performance testing `conv_separable()`
        Usage:
            conv [options]

        Options:
            -d DIMS  : Up to [default: 3:int]
            -k WIDTH  : Up to [default: 17:int]
            -i WIDTH  : input width [default: 234:int]
        """
        assert 0 < args.d <= 3
        assert 0 < args.k <= 17
        KNL = cu.asarray(np.random.random((args.d, args.k)), dtype='float32')
        SRC = cu.asarray(np.random.random((args.i,) * args.d), dtype='float32')
        for _ in trange(args.repeats, unit="repeats",
                        desc=f"{args.i}^{args.d} (*) {args.k}^{args.d}"):
            dst_gpu = prc.conv_separable(SRC, KNL)
        assert hasattr(dst_gpu, 'cuvec') or args.d < 3

    argopt(dedent(conv.__doc__), argparser=sub_parser).set_defaults(func=conv)

    def nlm(args):
        """\
        Performance testing `nlm()`
        Usage:
            nlm [options]

        Options:
            -s SIGMA  : NLM parameter [default: 1:float]
            -k HALF_WIDTH  : kernel half-width [default: 2:int]
            -i WIDTH  : input width [default: 123:int]
        """
        assert 0 < args.k
        assert 0 < args.i
        IMG = cu.asarray(np.random.random((args.i,) * 3), dtype='float32')
        REF = cu.asarray(np.random.random((args.i,) * 3), dtype='float32')
        for _ in trange(args.repeats, unit="repeats", desc=f"{args.i}^3 (*) {args.k*2+1}^3"):
            dst_gpu = prc.nlm(IMG, REF, sigma=args.s, half_width=args.k)
        assert hasattr(dst_gpu, 'cuvec')

    argopt(dedent(nlm.__doc__), argparser=sub_parser).set_defaults(func=nlm)

    args = parser.parse_args()
    if hasattr(args, 'func'):
        args.func(args)
    else:     # py<=3.6
        parser.parse_args(['-h'])



"""
NIMPA: functions for neuro image processing and analysis.
Includes functions relating to image registration/segmentation.
"""
import glob
import logging
import os
import shutil
import sys
from subprocess import call
from textwrap import dedent

import numpy as np
import scipy.ndimage as ndi
from miutil.fdio import hasext
from spm12.regseg import resample_spm  # NOQA: F401 yapf: disable
from spm12.regseg import coreg_spm

from .. import resources as rs
from . import imio, prc

log = logging.getLogger(__name__)


def imfill(immsk):
    '''fill the empty patches of image mask 'immsk' '''
    for iz in range(immsk.shape[0]):
        for iy in range(immsk.shape[1]):
            ix0 = np.argmax(immsk[iz, iy, :] > 0)
            ix1 = immsk.shape[2] - np.argmax(immsk[iz, iy, ::-1] > 0)
            if (ix1 - ix0) > immsk.shape[2] - 10: continue
            immsk[iz, iy, ix0:ix1] = 1
    return immsk


# ------------------------------------------------------------------------------
# Create object mask for the input image
# ------------------------------------------------------------------------------


def create_mask(
    fnii,
    fimout='',
    outpath='',
    fill=1,
    dtype_fill=np.uint8,
    thrsh=0.,
    fwhm=0.,
):
    '''create mask over the whole image or over the threshold area'''
    # > output path
    if outpath == '' and fimout != '':
        opth = os.path.dirname(fimout)
        if opth == '':
            opth = os.path.dirname(fnii)
            fimout = os.path.join(opth, fimout)

    elif outpath == '':
        opth = os.path.dirname(fnii)

    else:
        opth = outpath

    # > output file name if not given
    if fimout == '':
        fniis = os.path.split(fnii)
        fimout = os.path.join(opth, fniis[1].split('.nii')[0] + '_mask.nii.gz')

    niidct = imio.getnii(fnii, output='all')
    im = niidct['im']
    hdr = niidct['hdr']

    if im.ndim > 3:
        raise ValueError('The masking function only accepts 3-D images.')

    # > generate output image
    if thrsh > 0.:
        smoim = ndi.filters.gaussian_filter(im, imio.fwhm2sig(fwhm, voxsize=abs(hdr['pixdim'][1])),
                                            mode='mirror')
        thrsh = thrsh * smoim.max()
        immsk = np.int8(smoim > thrsh)
        immsk = imfill(immsk)

        # > output image
        imo = fill * immsk.astype(dtype_fill)
    else:
        imo = fill * np.ones(im.shape, dtype=dtype_fill)

    # > save output image
    imio.array2nii(
        imo, niidct['affine'], fimout,
        trnsp=(niidct['transpose'].index(0), niidct['transpose'].index(1),
               niidct['transpose'].index(2)), flip=niidct['flip'])

    return {'fim': fimout, 'im': imo}


# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
# I M A G E   R E G I S T R A T I O N
# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>


def affine_niftyreg(
    fref,
    fflo,
    outpath='',
    fname_aff='',
    pickname='ref',
    fcomment='',
    executable='',
    omp=1,
    rigOnly=False,
    affDirect=False,
    maxit=5,
    speed=True,
    pi=50,
    pv=50,
    smof=0,
    smor=0,
    rmsk=True,
    fmsk=True,
    rfwhm=15.,
    rthrsh=0.05,
    ffwhm=15.,
    fthrsh=0.05,
    verbose=True,
):

    # check if the executable exists:
    if not os.path.isfile(executable):
        raise IOError('Incorrect path to executable file for registration.')

    # create a folder for images registered to ref
    if outpath != '':
        odir = os.path.join(outpath, 'affine-niftyreg')
        fimdir = os.path.join(outpath, os.path.join('affine-niftyreg', 'mask'))
    else:
        odir = os.path.join(os.path.dirname(fflo), 'affine-niftyreg')
        fimdir = os.path.join(os.path.dirname(fflo), 'affine-niftyreg', 'mask')
    imio.create_dir(odir)
    imio.create_dir(fimdir)

    if rmsk:
        f_rmsk = os.path.join(fimdir,
                              'rmask_' + os.path.basename(fref).split('.nii')[0] + '.nii.gz')
        create_mask(fref, fimout=f_rmsk, thrsh=rthrsh, fwhm=rfwhm)

    if fmsk:
        f_fmsk = os.path.join(fimdir,
                              'fmask_' + os.path.basename(fflo).split('.nii')[0] + '.nii.gz')
        create_mask(fflo, fimout=f_fmsk, thrsh=fthrsh, fwhm=ffwhm)

    # output in register with ref and text file for the affine transform
    if fname_aff != '':
        fout = os.path.join(odir, fname_aff.split('.')[0] + '.nii.gz')
        faff = os.path.join(odir, fname_aff)
    else:
        if pickname == 'ref':
            fout = os.path.join(
                odir,
                'affine_ref-' + os.path.basename(fref).split('.nii')[0] + fcomment + '.nii.gz')
            faff = os.path.join(
                odir, 'affine_ref-' + os.path.basename(fref).split('.nii')[0] + fcomment + '.txt')
        elif pickname == 'flo':
            fout = os.path.join(
                odir,
                'affine_flo-' + os.path.basename(fflo).split('.nii')[0] + fcomment + '.nii.gz')
            faff = os.path.join(
                odir, 'affine_flo-' + os.path.basename(fflo).split('.nii')[0] + fcomment + '.txt')

    # call the registration routine
    cmd = [
        executable, '-ref', fref, '-flo', fflo, '-aff', faff, '-pi',
        str(pi), '-pv',
        str(pv), '-smooF',
        str(smof), '-smooR',
        str(smor), '-maxit', '10', '-omp',
        str(omp), '-res', fout]
    if speed:
        cmd.append('-speeeeed')
    if rigOnly:
        cmd.append('-rigOnly')
    if affDirect:
        cmd.append('affDirect')
    if rmsk:
        cmd.append('-rmask')
        cmd.append(f_rmsk)
    if fmsk:
        cmd.append('-fmask')
        cmd.append(f_fmsk)
    if not verbose:
        cmd.append('-voff')

    call(cmd)

    # > affine to Numpy array
    aff = np.loadtxt(faff)

    return {'affine': aff, 'faff': faff, 'fim': fout} # faff, fout


def resample_niftyreg(
    fref,
    fflo,
    faff,
    outpath='',
    fimout='',
    fcomment='',
    pickname='ref',
    intrp=1,
    executable='',
    verbose=True,
):

    # check if the executable exists:
    # if executable=='' and 'RESPATH' in Cnt and os.path.isfile(Cnt['RESPATH']):
    #     executable = Cnt['RESPATH']

    if not os.path.isfile(executable):
        raise IOError('Incorrect path to executable file for registration.')

    # > output path
    if outpath == '' and fimout != '':
        opth = os.path.dirname(fimout)
        if opth == '':
            opth = os.path.dirname(fflo)
    elif outpath == '':
        opth = os.path.dirname(fflo)
    else:
        opth = outpath

    imio.create_dir(opth)

    # > the output naming
    if '/' in fimout:
        fout = fimout
    elif fimout != '' and not os.path.isfile(fimout):
        fout = os.path.join(opth, fimout)
    elif pickname == 'ref':
        fout = os.path.join(
            opth, 'affine_ref-' + os.path.basename(fref).split('.nii')[0] + fcomment + '.nii.gz')
    elif pickname == 'flo':
        fout = os.path.join(
            opth, 'affine_flo-' + os.path.basename(fflo).split('.nii')[0] + fcomment + '.nii.gz')

    if isinstance(intrp, int):
        intrp = str(intrp)

    cmd = [executable, '-ref', fref, '-flo', fflo, '-trans', faff, '-res', fout, '-inter', intrp]
    if not verbose:
        cmd.append('-voff')
    call(cmd)

    return fout


# ==============================================================================
# S P M realignment of multiple images through m-scripting (dynamic PET)
# ==============================================================================


def realign_mltp_spm(
    fims,
    quality=1.0,
    fwhm=6,
    sep=4,
    rtm=1,
    interp=2,
    graph=0,
    outpath='',
    fcomment='',
    niicopy=False,
    niisort=False,
):
    '''
    Arguments:
        fims:   has to be a list of at least two files with the first one acting
                as a reference.
    '''
    # > input folder
    inpath = os.path.dirname(fims[0])

    # > output folder
    if outpath == '':
        outpath = os.path.join(inpath, 'align')
    else:
        outpath = os.path.join(outpath, 'align')

    if fims[0][-3:] == '.gz' or niicopy:
        tmpth = outpath # os.path.join(outpath, 'tmp')
        rpth = tmpth
    else:
        tmpth = outpath
        rpth = inpath

    imio.create_dir(tmpth)

    # > uncompress for SPM
    fungz = []
    for f in fims:
        if f[-3:] == '.gz':
            fun = imio.nii_ugzip(f, outpath=tmpth)
        elif os.path.isfile(f) and f.endswith('nii'):
            if niicopy:
                fun = os.path.join(tmpth, os.path.basename(f).split('.nii')[0] + '_copy.nii')
                shutil.copyfile(f, fun)
            else:
                fun = f
        else:
            log.warning('omitting file/folder: {}'.format(f))
        fungz.append(fun)

    if niisort:
        niisrt = imio.niisort([
            os.path.join(tmpth, f) for f in os.listdir(tmpth)
            if os.path.isfile(os.path.join(tmpth, f))])
        fsrt = niisrt['files']
    else:
        fsrt = fungz

    P_input = [f + ',1' for f in fsrt if f.endswith('nii') and f[0] != 'r' and 'mean' not in f]

    # > maximal number of characters per line (for Matlab array)
    Pinmx = max(map(len, P_input))

    # > equalise the input char array
    Pineq = [f.ljust(Pinmx) for f in P_input]

    # --------------------------------------------------------------------------
    # > MATLAB realign flags for SPM
    flgs = []
    pw = '\'\''

    flgs.append('flgs.quality = ' + str(quality))
    flgs.append('flgs.fwhm = ' + str(fwhm))
    flgs.append('flgs.sep = ' + str(sep))
    flgs.append('flgs.rtm = ' + str(rtm))
    flgs.append('flgs.pw  = ' + str(pw))
    flgs.append('flgs.interp = ' + str(interp))
    flgs.append('flgs.graphics = ' + str(graph))
    flgs.append('flgs.wrap = [0 0 0]')

    flgs.append('disp(flgs)')
    # --------------------------------------------------------------------------

    fscript = os.path.join(outpath, 'pyauto_script_realign.m')

    with open(fscript, 'w') as f:
        f.write('% AUTOMATICALLY GENERATED MATLAB SCRIPT FOR SPM PET REALIGNMENT\n\n')

        f.write('%> the following flags for PET image alignment are used:\n')
        f.write("disp('m> SPM realign flags:');\n")
        for item in flgs:
            f.write("%s;\n" % item)

        f.write('\n%> the following PET images will be aligned:\n')
        f.write("disp('m> PET images for SPM realignment:');\n")

        f.write('P{1,1} = [...\n')
        for item in Pineq:
            f.write("'%s'\n" % item)
        f.write('];\n\n')
        f.write('disp(P{1,1});\n')

        f.write('spm_realign(P, flgs);')

    cmd = ['matlab', '-nodisplay', '-nodesktop', '-r', 'run(' + '\'' + fscript + '\'' + '); exit']
    call(cmd)

    fres = glob.glob(os.path.join(rpth, 'rp*.txt'))[0]
    res = np.loadtxt(fres)

    return {'fout': fres, 'outarray': res, 'P': Pineq, 'fims': fsrt}


# ==============================================================================
# S P M resampling of multiple images through m-scripting (dynamic PET)
# ==============================================================================


def resample_mltp_spm(
    fims,
    ftr,
    interp=1.,
    which=1,
    mask=0,
    mean=0,
    graph=0,
    niisort=False,
    prefix='r_',
    outpath='',
    fcomment='',
    pickname='ref',
    copy_input=False,
    del_in_uncmpr=False,
    del_out_uncmpr=False,
):
    '''
    Arguments:
        fims:   has to be a list of at least two files with the first one acting
                as a reference.
    '''
    if not isinstance(fims, list) and not isinstance(fims[0], str):
        raise ValueError('e> unrecognised list of input images')

    if not os.path.isfile(ftr):
        raise IOError('e> cannot open the file with translations and rotations')

    # > output path
    if outpath == '':
        opth = os.path.dirname(fims[0])
    else:
        opth = outpath

    imio.create_dir(opth)

    # > working file names (not touching the original ones)
    _fims = []

    # > decompress if necessary
    for f in fims:
        if not os.path.isfile(f) and not (f.endswith('nii') or f.endswith('nii.gz')):
            raise IOError('e> could not open file:', f)

        if f[-3:] == '.gz':
            fugz = imio.nii_ugzip(f, outpath=os.path.join(opth, 'copy'))
        elif copy_input:
            fnm = os.path.basename(f).split('.nii')[0] + '_copy.nii'
            fugz = os.path.join(opth, 'copy', fnm)
            shutil.copyfile(f, fugz)
        else:
            fugz = f

        _fims.append(fugz)

    if niisort:
        niisrt = imio.niisort(_fims)
        _fims = niisrt['files']

    # > maximal number of characters per line (for Matlab array)
    Pinmx = max(map(len, _fims))

    # > equalise the input char array
    Pineq = [f.ljust(Pinmx) for f in _fims]

    # --------------------------------------------------------------------------
    # > SPM reslicing (resampling) flags
    flgs = []

    flgs.append('flgs.mask = ' + str(mask))
    flgs.append('flgs.mean = ' + str(mean))

    flgs.append('flgs.which = ' + str(which))
    flgs.append('flgs.interp = ' + str(interp))
    flgs.append('flgs.graphics = ' + str(graph))
    flgs.append('flgs.wrap = [0 0 0]')
    flgs.append('flgs.prefix = ' + "'" + prefix + "'")

    flgs.append('disp(flgs)')
    # --------------------------------------------------------------------------

    fsrpt = os.path.join(opth, 'pyauto_script_resampling.m')
    with open(fsrpt, 'w') as f:
        f.write('% AUTOMATICALLY GENERATED MATLAB SCRIPT FOR SPM RESAMPLING PET IMAGES\n\n')

        f.write('%> the following flags for image reslicing are used:\n')
        f.write("disp('m> SPM reslicing flags:');\n")
        for item in flgs:
            f.write("%s;\n" % item)

        f.write('\n%> the following PET images will be aligned:\n')
        f.write("disp('m> PET images for SPM reslicing:');\n")

        f.write('P{1,1} = [...\n')
        for item in Pineq:
            f.write("'%s'\n" % item)
        f.write('];\n\n')
        f.write('disp(P{1,1});\n')

        # f.write('\n%> the following PET images will be aligned using'
        #         ' the translations and rotations in X:\n')
        # f.write("X = dlmread('"+ftr+"');\n")
        # f.write('for fi = 2:'+str(len(fims))+'\n')
        # f.write("    VF = strcat(P{1,1}(fi,:),',1');\n")
        # f.write('    M_ = zeros(4,4);\n')
        # f.write('    M_(:,:) = spm_get_space(VF);\n')
        # f.write('    M = spm_matrix(X(fi,:));\n')
        # f.write('    spm_get_space(VF, M\M_(:,:));\n')
        # f.write('end\n\n')

        f.write('spm_reslice(P, flgs);\n')

    cmd = ['matlab', '-nodisplay', '-nodesktop', '-r', 'run(' + '\'' + fsrpt + '\'' + '); exit']
    call(cmd)

    if del_in_uncmpr:
        for fpth in _fims:
            os.remove(fpth)


# ==============================================================================
# V I N C I  registration and resampling
# ==============================================================================


def coreg_vinci(
    fref,
    fflo,
    vc='',
    con='',
    vincipy_path='',
    scheme_xml='',
    outpath='',
    fname_aff='',
    pickname='ref',
    fcomment='',
    flo_colourmap='Green',
    close_vinci=False,
    close_buff=True,
    cleanup=True,
    save_res=False,
):
    if scheme_xml == '':
        raise IOError('e> the Vinci schema *.xml file is not provided. \n \
                i> please add the schema file in the call: scheme_xml=...')

    # --------------------------------------------------------------------------
    # > output path
    if outpath == '' and fname_aff != '' and '/' in fname_aff:
        opth = os.path.dirname(fname_aff)
        if opth == '':
            opth = os.path.dirname(fflo)
        fname_aff = os.path.basename(fname_aff)

    elif outpath == '':
        opth = os.path.dirname(fflo)
    else:
        opth = outpath
    imio.create_dir(opth)

    imio.create_dir(os.path.join(opth, 'affine-vinci'))

    # > output floating and affine file names
    if fname_aff == '':
        if pickname == 'ref':

            faff = os.path.join(
                opth, 'affine-vinci',
                'affine-ref-' + os.path.basename(fref).split('.nii')[0] + fcomment + '.xml')

            fout = os.path.join(
                opth, 'affine-vinci',
                'affine-ref-' + os.path.basename(fref).split('.nii')[0] + fcomment + '.nii.gz')

        else:
            faff = os.path.join(
                opth, 'affine-vinci',
                'affine-flo-' + os.path.basename(fflo).split('.nii')[0] + fcomment + '.xml')

            fout = os.path.join(
                opth, 'affine-vinci',
                'affine-flo-' + os.path.basename(fflo).split('.nii')[0] + fcomment + '.nii.gz')

    else:

        # > add '.xml' extension if not in the affine output file name
        if not fname_aff.endswith('.xml'):
            fname_aff += '.xml'

        faff = os.path.join(opth, 'affine-vinci', fname_aff)

        fout = os.path.join(opth, fname_aff.split('.')[0] + '.nii.gz')
    # --------------------------------------------------------------------------

    if vincipy_path == '':
        try:
            vincipy_path = rs.VINCIPATH
        except AttributeError:
            raise AttributeError('e> could not find variable VINCIPATH in resources.py')

    sys.path.append(vincipy_path)

    try:
        from VinciPy import Vinci_Bin, Vinci_Connect, Vinci_Core, Vinci_ImageT, Vinci_XML
    except ImportError:
        raise ImportError('e> could not import Vinci:\n \
                check the variable VINCIPATH (path to Vinci) in resources.py')

    # > start Vinci core engine if it is not running/given
    if vc == '' or con == '':
        # > adopted from the Vinci's example RunMMMJob.py
        bin = Vinci_Bin.Vinci_Bin()
        con = Vinci_Connect.Vinci_Connect(bin)
        con.StartMyVinci()

        vc = Vinci_Core.Vinci_CoreCalc(con)
        vc.StdProject()

    # > read the registration schema file
    f = open(scheme_xml, 'rb')
    reg_scheme = f.read()
    f.close()

    # > check scheme data file and remove possible XML comments at its beginning
    root = Vinci_XML.ElementTree.fromstring(reg_scheme)
    if root.tag != "MMM":
        sys.exit("scheme data file %s does not contain tag MMM\n" % scheme_xml)
    reg_scheme = Vinci_XML.ElementTree.tostring(root)
    # bytes -> str
    reg_scheme = reg_scheme.decode("utf-8")

    # > pick reference and floating images
    ref = Vinci_ImageT.newTemporary(vc, szFileName=fref)
    rsl = Vinci_ImageT.newTemporary(vc, szFileName=fflo)

    # > set the colour map for the floating image
    rsl.setColorSettings(CTable=flo_colourmap)

    # --------------------------------------------------------------------------
    # > perform the registration
    rsl.alignToRef(ref, reg_scheme, szRegistrationSummaryFile=faff)
    # --------------------------------------------------------------------------

    # > find the weird folder name created and to be disposed of.
    if cleanup:
        # > get the folder name (going to rubbish)
        del_dir = os.path.normpath(faff).split(os.sep)[1]
        del_dir = os.path.join(os.path.split(fflo)[0], os.path.normpath(faff).split(os.sep)[1])
        if os.path.isdir(del_dir):
            shutil.rmtree(del_dir)

    # > save the registered image
    if save_res:
        rsl.saveYourselfAs(bUseOffsetRotation=True, szFullFileName=fout)

    # > close image buffers for reference and floating
    if close_buff:
        rsl.killYourself()
        ref.killYourself()

    if close_vinci: con.CloseVinci(True)

    return {'faff': faff, 'fim': fout, 'vinci_con': con, 'vinci_vc': vc}


# ------------------------------------------------------------------------------
# VINCI RESAMPLE
# ------------------------------------------------------------------------------


def resample_vinci(
    fref,
    fflo,
    faff,
    intrp=0,
    fimout='',
    fcomment='',
    outpath='',
    pickname='ref',
    vc='',
    con='',
    vincipy_path='',
    atlas_resample=False,
    atlas_ref_make=False,
    atlas_ref_del=False,
    close_vinci=False,
    close_buff=True,
):
    '''
    Resample the floating image <fflo> into the geometry of <fref>,
    using the Vinci transformation output <faff> (an *.xml file).
    Output the NIfTI file path of the resampled/resliced image.
    '''
    # > output path
    if outpath == '' and fimout != '' and '/' in fimout:
        opth = os.path.dirname(fimout)
        if opth == '':
            opth = os.path.dirname(fflo)
        fimout = os.path.basename(fimout)

    elif outpath == '':
        opth = os.path.dirname(fflo)
    else:
        opth = outpath
    imio.create_dir(opth)

    # > output floating and affine file names
    if fimout == '':
        if pickname == 'ref':
            fout = os.path.join(
                opth,
                'affine-ref-' + os.path.basename(fref).split('.nii')[0] + fcomment + '.nii.gz')
        else:
            fout = os.path.join(
                opth,
                'affine-flo-' + os.path.basename(fflo).split('.nii')[0] + fcomment + '.nii.gz')
    else:
        fout = os.path.join(opth, fimout.split('.')[0] + '.nii.gz')
    # --------------------------------------------------------------------------

    if vincipy_path == '':
        try:
            vincipy_path = rs.VINCIPATH
        except AttributeError:
            raise AttributeError('e> could not find variable VINCIPATH in resources.py')

    sys.path.append(vincipy_path)

    try:
        from VinciPy import Vinci_Bin, Vinci_Connect, Vinci_Core, Vinci_ImageT
    except ImportError:
        raise ImportError('e> could not import Vinci:\n \
                check the variable VINCIPATH (path to Vinci) in resources.py')

    # --------------------------------------------------------------------------
    # > start Vinci core engine if it is not running/given
    if vc == '' or con == '':
        # > adopted from the Vinci's example RunMMMJob.py
        bin = Vinci_Bin.Vinci_Bin()
        con = Vinci_Connect.Vinci_Connect(bin)
        con.StartMyVinci()

        vc = Vinci_Core.Vinci_CoreCalc(con)
        vc.StdProject()
    # --------------------------------------------------------------------------

    if int(intrp) == 0:
        interp_nn = True
    else:
        interp_nn = False

    # --------------------------------------------------------------------------
    # > change the reference image to be loaded as atlas
    if atlas_resample and atlas_ref_make:

        # > output file name for the extra reference image
        fextref = os.path.join(opth, 'reference-as-atlas.nii.gz')

        prc.nii_modify(fref, fimout=fextref, voxel_range=[0., 255.])

        ref = Vinci_ImageT.newTemporary(vc, szFileName=fextref, bLoadAsAtlas=atlas_resample)

        if atlas_ref_del:
            os.remove(fextref)

    else:
        ref = Vinci_ImageT.newTemporary(vc, szFileName=fref, bLoadAsAtlas=atlas_resample)
    # --------------------------------------------------------------------------

    # --------------------------------------------------------------------------
    # > reapply rigid transformation to new image

    flo = Vinci_ImageT.newTemporary(vc, szFileName=fflo, bLoadAsAtlas=atlas_resample)

    rsl = flo.reapplyMMMTransform(faff, refImage=ref, IsComputed=True)

    rsl.saveYourselfAs(bUseOffsetRotation=True, bUseNextNeighbourInterp=interp_nn,
                       szFullFileName=fout)
    # --------------------------------------------------------------------------

    # > close image buffers for reference and floating
    if close_buff:
        ref.killYourself()
        flo.killYourself()
        rsl.killYourself()

    if close_vinci: con.CloseVinci(True)

    return fout


def affine_fsl(
    fref,
    fflo,
    outpath='',
    fname_aff='',
    pickname='ref',
    fcomment='',
    executable='',
    costfun='normmi',
    dof=6,
    hstbins=256,
    verbose=True,
):
    """
    FSL-FLIRT REGISTRATION (AFFINE/RIGID)
    """
    if executable == '':
        if 'FSLDIR' not in os.environ:
            raise IOError('e> no FSL executable provided!')
        else:
            executable = os.path.join(os.environ['FSLDIR'], 'bin', 'flirt')

    if not os.path.isfile(executable) or not call([executable, '-version']) == 0:
        raise IOError('e> no valid FSL executable provided!')

    # --------------------------------------------------------------------------
    # > output path
    if outpath == '' and fname_aff != '' and os.path.isfile(fname_aff):
        opth = os.path.dirname(fname_aff)
        if opth == '':
            opth = os.path.dirname(fflo)
        fname_aff = os.path.basename(fname_aff)

    elif outpath == '':
        opth = os.path.dirname(fflo)
    else:
        opth = outpath
    imio.create_dir(opth)

    imio.create_dir(os.path.join(opth, 'affine-fsl'))

    # > output floating and affine file names
    if fname_aff == '':

        if pickname == 'ref':
            faff = os.path.join(
                opth, 'affine-fsl',
                'affine-ref-' + os.path.basename(fref).split('.nii')[0] + fcomment + '.txt')

        else:
            faff = os.path.join(
                opth, 'affine-fsl',
                'affine-flo-' + os.path.basename(fflo).split('.nii')[0] + fcomment + '.txt')

    else:

        # > add '.xml' extension if not in the affine output file name
        if not fname_aff.endswith('.txt'):
            fname_aff += '.txt'

        faff = os.path.join(opth, 'affine-vinci', fname_aff)
    # --------------------------------------------------------------------------

    # > command with options for FSL-FLIRT registration
    cmd = [
        executable, '-cost', costfun, '-dof',
        str(dof), '-omat', faff, '-in', fflo, '-ref', fref, '-bins',
        str(hstbins)]

    # > add verbose mode if requested
    if isinstance(verbose, bool): verbose = int(verbose)
    if verbose > 0: cmd.extend(['-verbose', str(verbose)])

    # > execute FSL-FLIRT command with the above options
    call(cmd)

    # convert to Numpy float
    aff = np.loadtxt(faff)

    return {'affine': aff, 'faff': faff}


# ------------------------------------------------------------------------------
# FSL RESAMPLING
# ------------------------------------------------------------------------------


def resample_fsl(
    imref,
    imflo,
    faff,
    outpath='',
    fimout='',
    pickname='ref',
    fcomment='',
    intrp=1,
    executable='',
):
    if executable == '':
        if 'FSLDIR' not in os.environ:
            raise IOError('e> no FSL executable provided!')
        else:
            executable = os.path.join(os.environ['FSLDIR'], 'bin', 'flirt')

    if not os.path.isfile(executable) or not call([executable, '-version']) == 0:
        raise IOError('e> no valid FSL executable provided!')

    log.info(
        dedent('''\
        ===================================================================
        F S L  resampling...
        ==================================================================='''))

    # > output path
    if outpath == '' and fimout != '':
        opth = os.path.dirname(fimout)
        if opth == '':
            opth = os.path.dirname(imflo)

    elif outpath == '':
        opth = os.path.dirname(imflo)
    elif outpath != '':
        opth = outpath

    imio.create_dir(opth)

    # > the output naming
    if '/' in fimout:
        fout = fimout
    elif fimout != '' and not os.path.isfile(fimout):
        fout = os.path.join(opth, fimout)
    elif pickname == 'ref':
        fout = os.path.join(
            opth, 'affine_ref-' + os.path.basename(imref).split('.nii')[0] + fcomment + '.nii.gz')
    elif pickname == 'flo':
        fout = os.path.join(
            opth, 'affine_flo-' + os.path.basename(imflo).split('.nii')[0] + fcomment + '.nii.gz')

    intrp = int(intrp)
    if isinstance(intrp, int): intrp = str(intrp)
    if intrp == '1':
        interpolation = 'trilinear'
    elif intrp == '0':
        interpolation = 'nearestneighbour'

    cmd = [
        executable, '-in', imflo, '-ref', imref, '-out', fout, '-init', faff, '-applyxfm',
        '-interp', interpolation]
    call(cmd)

    log.info(
        dedent('''\
        ===================================================================
        F S L  resampling done.
        ==================================================================='''))

    return fout


# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
# M O T I O N
# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>


def motion_reg(
    ref,
    flo,
    fcomment='',
    pickname='ref',
    outpath='',
    fname_aff='',
    matlab_eng_name='',
    rot_thresh=1.,
    trn_thresh=1.,
):
    if isinstance(flo, str):
        flolst = [flo]
    elif isinstance(flo, list) and all(os.path.isfile(f) for f in flo):
        flolst = flo
    else:
        raise OSError('could not decode the input of floating images.')

    # dctout = {}
    lstout = []

    motion_rot = False
    motion_trn = False

    for i in range(len(flolst)):
        M = coreg_spm(ref, flolst[i], fcomment=fcomment, pickname=pickname, fname_aff=fname_aff,
                      matlab_eng_name=matlab_eng_name, outpath=outpath, visual=0, del_uncmpr=True)

        if any((np.abs(M['rotations']) * 180 / np.pi) > rot_thresh):
            log.warning(
                dedent('''\
                at least one rotation is above the threshold of {}, and is:
                {}''').format(rot_thresh, M['rotations'] * 180 / np.pi))
            motion_rot = True
        if any(np.abs(M['translations']) > trn_thresh):
            log.warning(
                dedent('''\
                at least one translation is above the threshold of {}, and is:
                {}''').format(trn_thresh, M['translations']))
            motion_trn = True

        lstout.append({'regout': M, 'trans_mo': motion_trn, 'rotat_mo': motion_rot})

    return lstout


# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
# I M A G E   S I M I L A R I T Y
# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>


def dice_coeff(im1, im2, val=1):
    '''
    Calculate Dice score for parcellation images <im1> and <im2> and ROI value <val>.
    Input images can be given as:
        1. paths to NIfTI image files or as
        2. Numpy arrays.
    The ROI value can be given as:
        1. a single integer representing one ROI out of many in the parcellation
           images (<im1> and <im2>) or as
        2. a list of integers to form a composite ROI used for the association test.
    Outputs a float number representing the Dice score.
    '''
    if all(
            isinstance(i, str) and os.path.isfile(i) and hasext(i, ('nii', 'nii.gz'))
            for i in [im1, im2]):
        imn1 = imio.getnii(im1, output='image')
        imn2 = imio.getnii(im2, output='image')
    elif isinstance(im1, (np.ndarray, np.generic)) and isinstance(im1, (np.ndarray, np.generic)):
        imn1 = im1
        imn2 = im2
    else:
        raise TypeError('Unrecognised or Mismatched Images.')

    # a single value corresponding to one ROI
    if isinstance(val, int):
        imv1 = (imn1 == val)
        imv2 = (imn2 == val)
    # multiple values in list corresponding to a composite ROI
    elif isinstance(val, list) and all(isinstance(v, int) for v in val):
        imv1 = (imn1 == val[0])
        imv2 = (imn2 == val[0])
        for v in val[1:]:
            # boolean addition to form a composite ROI
            imv1 += (imn1 == v)
            imv2 += (imn2 == v)
    else:
        raise TypeError('ROI Values have to be integer (single or in a list).')

    if imv1.shape != imv2.shape:
        raise ValueError('Shape Mismatch: Input images must have the same shape.')

    # compute Dice coefficient
    intrsctn = np.logical_and(imv1, imv2)

    return 2. * intrsctn.sum() / (imv1.sum() + imv2.sum())


def dice_coeff_multiclass(im1, im2, roi2ind):
    '''
    Calculate Dice score for parcellation images <im1> and <im2> and ROI value <val>.
    Input images can be given as:
        1. paths to NIfTI image files or as
        2. Numpy arrays.
    The ROI value must be given as a dictionary of lists of indexes for each ROI
    Outputs a float number representing the Dice score.
    '''
    if all(
            isinstance(i, str) and os.path.isfile(i) and hasext(i, ('nii', 'nii.gz'))
            for i in [im1, im2]):
        imn1 = imio.getnii(im1, output='image')
        imn2 = imio.getnii(im2, output='image')
    elif isinstance(im1, (np.ndarray, np.generic)) and isinstance(im1, (np.ndarray, np.generic)):
        imn1 = im1
        imn2 = im2
    else:
        raise TypeError('Unrecognised or Mismatched Images.')

    if imn1.shape != imn2.shape:
        raise ValueError('Shape Mismatch: Input images must have the same shape.')

    out = {}
    for k in roi2ind:

        # multiple values in list corresponding to a composite ROI
        imv1 = (imn1 == roi2ind[k][0])
        imv2 = (imn2 == roi2ind[k][0])
        for v in roi2ind[k][1:]:
            # boolean addition to form a composite ROI
            imv1 += (imn1 == v)
            imv2 += (imn2 == v)

        # compute Dice coefficient
        intrsctn = np.logical_and(imv1, imv2)
        out[k] = 2. * intrsctn.sum() / (imv1.sum() + imv2.sum())

    return out



from os import path

from niftypet import nimpa


def test_dev_info(capsys):
    devs = nimpa.dev_info()
    out, err = capsys.readouterr()
    assert not any((out, err))
    nimpa.dev_info(showprop=True)
    out, err = capsys.readouterr()
    assert not err
    assert not devs or out


def test_resources():
    assert path.exists(nimpa.path_resources)
    assert nimpa.resources.DIRTOOLS



"""image input/output functionalities."""
import datetime
import glob
import logging
import os
import re
import shutil
from subprocess import run
from textwrap import dedent

import nibabel as nib
import numpy as np
import pydicom as dcm

# > NiftyPET resources
from .. import resources as rs

log = logging.getLogger(__name__)

# possible extentions for DICOM files
dcmext = ('dcm', 'DCM', 'ima', 'IMA')


def create_dir(pth):
    if not os.path.exists(pth):
        os.makedirs(pth)


def time_stamp(simple_ascii=False):
    now = datetime.datetime.now()
    if simple_ascii:
        nowstr = str(now.year) + '-' + str(now.month) + '-' + str(now.day) + '_' + str(
            now.hour) + 'h' + str(now.minute)
    else:
        nowstr = str(now.year) + '-' + str(now.month) + '-' + str(now.day) + ' ' + str(
            now.hour) + ':' + str(now.minute)
    return nowstr


def fwhm2sig(fwhm, voxsize=2.0):
    return (fwhm/voxsize) / (2 * (2 * np.log(2))**.5)


def getnii(fim, nan_replace=None, output='image'):
    '''
    Get PET image from NIfTI file.
    Arguments:
        fim: input file name for the nifty image
        nan_replace: the value to be used for replacing the NaNs in the image.
                     by default no change (None).
        output: option for choosing output: image, affine matrix or
                a dictionary with all info.
    Return:
        'image': outputs just an image (4D or 3D)
        'affine': outputs just the affine matrix
        'all': outputs all as a dictionary
    '''
    import numbers

    nim = nib.load(fim)

    dim = nim.header.get('dim')
    dimno = dim[0]

    if output == 'image' or output == 'all':
        imr = np.asanyarray(nim.dataobj)
        # replace NaNs if requested
        if isinstance(nan_replace, numbers.Number):
            imr[np.isnan(imr)] = nan_replace

        imr = np.squeeze(imr)
        if dimno != imr.ndim and dimno == 4:
            dimno = imr.ndim

        # > get orientations from the affine
        ornt = nib.io_orientation(nim.affine)
        trnsp = tuple(2 - np.int8(ornt[:, 0]))
        flip = tuple(np.int8(ornt[:, 1]))

        # > voxel size
        voxsize = nim.header.get('pixdim')[1:nim.header.get('dim')[0] + 1]
        # > rearrange voxel size according to the orientation
        voxsize = voxsize[np.array(trnsp)]

        # > dimensions
        dims = dim[1:nim.header.get('dim')[0] + 1]
        dims = dims[np.array(trnsp)]

        # > flip y-axis and z-axis and then transpose
        if dimno == 4:   # dynamic
            imr = np.transpose(imr[::-flip[0], ::-flip[1], ::-flip[2], :], (3,) + trnsp)
        elif dimno == 3: # static
            imr = np.transpose(imr[::-flip[0], ::-flip[1], ::-flip[2]], trnsp)

    if output == 'affine' or output == 'all':
        # A = nim.get_sform()
        # if not A[:3,:3].any():
        #     A = nim.get_qform()
        A = nim.affine

    if output == 'all':
        out = {
            'im': imr, 'affine': A, 'fim': fim, 'dtype': nim.get_data_dtype(), 'shape': imr.shape,
            'hdr': nim.header, 'voxsize': voxsize, 'dims': dims, 'transpose': trnsp, 'flip': flip}
    elif output == 'image':
        out = imr
    elif output == 'affine':
        out = A
    else:
        raise NameError('Unrecognised output request!')

    return out


def getnii_descr(fim):
    '''Extracts the custom description header field to dictionary'''
    nim = nib.load(fim)
    hdr = nim.header
    rcnlst = hdr['descrip'].item().split(';')
    rcndic = {}

    if rcnlst[0] == '':
        return rcndic

    for ci in range(len(rcnlst)):
        tmp = rcnlst[ci].split('=')
        rcndic[tmp[0]] = tmp[1]
    return rcndic


def array2nii(im, A, fnii, descrip='', trnsp=None, flip=None, storage_as=None):
    '''
    Store the numpy array 'im' to a NIfTI file 'fnii'.
    Arguments:
        'im':       image to be stored in NIfTI
        'A':        affine transformation
        'fnii':     output NIfTI file name.
        'descrip':  the description given to the file
        'trsnp':    transpose/permute the dimensions.
                    In NIfTI it has to be in this order: [x,y,z,t,...])
        'flip':     flip tupple for flipping the direction of x,y,z axes.
                    (1: no flip, -1: flip)
        'storage_as': uses the flip and displacement as given by the following
                    NifTI dictionary, obtained using
                    nimpa.getnii(filepath, output='all').
    '''

    if trnsp is None:
        trnsp = ()
    if flip is None:
        flip = ()
    if storage_as is None:
        storage_as = []

    if not len(trnsp) in [0, 3, 4] and not len(flip) in [0, 3]:
        raise ValueError('e> number of flip and/or transpose elements is incorrect.')

    # --------------------------------------------------------------------------
    # > TRANSLATIONS and FLIPS
    # > get the same geometry as the input NIfTI file in the form of dictionary,
    # >>as obtained from getnii(..., output='all')

    # > permute the axis order in the image array
    if isinstance(storage_as, dict) and 'transpose' in storage_as \
            and 'flip' in storage_as:

        trnsp = (storage_as['transpose'].index(0), storage_as['transpose'].index(1),
                 storage_as['transpose'].index(2))

        flip = storage_as['flip']

    if not trnsp:
        im = im.transpose()
    # > check if the image is 4D (dynamic) and modify as needed
    elif len(trnsp) == 3 and im.ndim == 4:
        trnsp = tuple([t + 1 for t in trnsp] + [0])
        im = im.transpose(trnsp)
    else:
        im = im.transpose(trnsp)

    # > perform flip of x,y,z axes after transposition into proper NIfTI order
    if flip and len(flip) == 3:
        im = im[::-flip[0], ::-flip[1], ::-flip[2], ...]
    # --------------------------------------------------------------------------

    nii = nib.Nifti1Image(im, A)
    hdr = nii.header
    hdr.set_sform(None, code='scanner')
    hdr['cal_max'] = np.max(im)
    hdr['cal_min'] = np.min(im)
    hdr['descrip'] = descrip
    nib.save(nii, fnii)


def orientnii(imfile, Cnt=None):
    '''Get the orientation from NIfTI sform.  Not fully functional yet.'''

    # > check if the dictionary of constant is given
    if Cnt is None:
        Cnt = {}

    strorient = ['L-R', 'S-I', 'A-P']
    niiorient = []
    niixyz = np.zeros(3, dtype=np.int8)
    if os.path.isfile(imfile):
        nim = nib.load(imfile)
        A = nim.get_sform()
        for i in range(3):
            niixyz[i] = np.argmax(abs(A[i, :-1]))
            niiorient.append(strorient[niixyz[i]])
        log.info('NIfTI orientation:\n{}'.format(niiorient))

    return niiorient


def nii_ugzip(imfile, outpath=''):
    '''Uncompress *.gz file'''
    import gzip
    with gzip.open(imfile, 'rb') as f:
        s = f.read()
    # Now store the uncompressed data
    if outpath == '':
        fout = imfile[:-3]
    else:
        fout = os.path.join(outpath, os.path.basename(imfile)[:-3])
    # store uncompressed file data from 's' variable
    with open(fout, 'wb') as f:
        f.write(s)
    return fout


def nii_gzip(imfile, outpath=''):
    '''Compress *.gz file'''
    import gzip
    with open(imfile, 'rb') as f:
        d = f.read()
    # Now store the compressed data
    if outpath == '':
        fout = imfile + '.gz'
    else:
        fout = os.path.join(outpath, os.path.basename(imfile) + '.gz')
    # store compressed file data from 'd' variable
    with gzip.open(fout, 'wb') as f:
        f.write(d)
    return fout


def pick_t1w(mri):
    '''Pick the MR T1w from the dictionary for MR->PET registration.'''

    if isinstance(mri, dict):
        # check if NIfTI file is given
        if 'T1N4' in mri and os.path.isfile(mri['T1N4']):
            ft1w = mri['T1N4']
        # or another bias corrected
        elif 'T1bc' in mri and os.path.isfile(mri['T1bc']):
            ft1w = mri['T1bc']
        elif 'T1nii' in mri and os.path.isfile(mri['T1nii']):
            ft1w = mri['T1nii']
        elif 'T1DCM' in mri and os.path.exists(mri['MRT1W']):
            # create file name for the converted NIfTI image
            fnii = 'converted'
            run([rs.DCM2NIIX, '-f', fnii, mri['T1nii']])
            ft1nii = glob.glob(os.path.join(mri['T1nii'], '*converted*.nii*'))
            ft1w = ft1nii[0]
        else:
            raise IOError('could not find a T1w image!')

    else:
        raise IOError('incorrect input given for the T1w image')

    return ft1w


def dcminfo(dcmvar, Cnt=None):
    '''Get basic info about the DICOM file/header.'''
    if Cnt is None:
        Cnt = {}

    if isinstance(dcmvar, str):
        log.debug('provided DICOM file: {}'.format(dcmvar))
        dhdr = dcm.dcmread(dcmvar)
    elif isinstance(dcmvar, dict):
        dhdr = dcmvar
    elif isinstance(dcmvar, dcm.dataset.FileDataset):
        dhdr = dcmvar

    dtype = dhdr[0x08, 0x08].value
    log.debug('   Image Type: {}'.format(dtype))

    # ------------------------------------------
    # > scanner ID
    scanner_vendor = 'unknown'
    if [0x008, 0x070] in dhdr:
        scanner_vendor = dhdr[0x008, 0x070].value

    scanner_model = 'unknown'
    if [0x008, 0x1090] in dhdr:
        scanner_model = dhdr[0x008, 0x1090].value

    scanner_id = 'other'
    if any(s in scanner_model
           for s in ['mMR', 'Biograph']) and 'siemens' in scanner_vendor.lower():
        scanner_id = 'mmr'
    # ------------------------------------------

    # > CSA type (mMR)
    csatype = ''
    if [0x29, 0x1108] in dhdr:
        csatype = dhdr[0x29, 0x1108].value
        log.debug('   CSA Data Type: {}'.format(csatype))

    # > DICOM comment or on MR parameters
    cmmnt = ''
    if [0x20, 0x4000] in dhdr:
        cmmnt = dhdr[0x0020, 0x4000].value
        log.debug('   Comments: {}'.format(cmmnt))

    # > MR parameters (echo time, etc)
    TR = 0
    TE = 0
    if [0x18, 0x80] in dhdr:
        TR = float(dhdr[0x18, 0x80].value)
        log.debug('   TR: {}'.format(TR))
    if [0x18, 0x81] in dhdr:
        TE = float(dhdr[0x18, 0x81].value)
        log.debug('   TE: {}'.format(TE))

    # > check if it is norm file
    if any('PET_NORM' in s
           for s in dtype) or cmmnt == 'PET Normalization data' or csatype == 'MRPETNORM':
        out = ['raw', 'norm', scanner_id]

    elif any('PET_LISTMODE' in s
             for s in dtype) or cmmnt == 'Listmode' or csatype == 'MRPETLM_LARGE':
        out = ['raw', 'list', scanner_id]

    elif any('MRPET_UMAP3D' in s for s in dtype) or cmmnt == 'MR based umap':
        out = ['raw', 'mumap', 'ute', 'mr', scanner_id]

    elif TR > 400 and TR < 2500 and TE < 20:
        out = ['mr', 't1', scanner_id]

    elif TR > 2500 and TE > 50:
        out = ['mr', 't2', scanner_id]

    # > UTE's two sequences: UTE2
    elif TR < 50 and TE < 20 and TE > 1:
        out = ['mr', 'ute', 'ute2', scanner_id]

    # > UTE1
    elif TR < 50 and TE < 20 and TE < 0.1 and TR > 0 and TE > 0:
        out = ['mr', 'ute', 'ute1', scanner_id]

    # > physio data
    elif 'PET_PHYSIO' in dtype or 'physio' in cmmnt.lower():
        out = ['raw', 'physio', scanner_id]

    else:
        out = ['unknown', str(cmmnt.lower())]

    return out


def list_dcm_datain(datain):
    '''List all DICOM file paths in the datain dictionary of input data.'''

    if not isinstance(datain, dict):
        raise ValueError('The input is not a dictionary!')

    dcmlst = []
    # list of mu-map DICOM files
    if 'mumapDCM' in datain:
        dcmump = os.listdir(datain['mumapDCM'])
        # accept only *.dcm extensions
        dcmump = [os.path.join(datain['mumapDCM'], d) for d in dcmump if d.endswith(dcmext)]
        dcmlst += dcmump

    if 'T1DCM' in datain:
        dcmt1 = os.listdir(datain['T1DCM'])
        # accept only *.dcm extensions
        dcmt1 = [os.path.join(datain['T1DCM'], d) for d in dcmt1 if d.endswith(dcmext)]
        dcmlst += dcmt1

    if 'T2DCM' in datain:
        dcmt2 = os.listdir(datain['T2DCM'])
        # accept only *.dcm extensions
        dcmt2 = [os.path.join(datain['T2DCM'], d) for d in dcmt2 if d.endswith(dcmext)]
        dcmlst += dcmt2

    if 'UTE1' in datain:
        dcmute1 = os.listdir(datain['UTE1'])
        # accept only *.dcm extensions
        dcmute1 = [os.path.join(datain['UTE1'], d) for d in dcmute1 if d.endswith(dcmext)]
        dcmlst += dcmute1

    if 'UTE2' in datain:
        dcmute2 = os.listdir(datain['UTE2'])
        # accept only *.dcm extensions
        dcmute2 = [os.path.join(datain['UTE2'], d) for d in dcmute2 if d.endswith(dcmext)]
        dcmlst += dcmute2

    # list-mode data dcm
    if 'lm_dcm' in datain:
        dcmlst += [datain['lm_dcm']]

    if 'lm_ima' in datain:
        dcmlst += [datain['lm_ima']]

    # norm
    if 'nrm_dcm' in datain:
        dcmlst += [datain['nrm_dcm']]

    if 'nrm_ima' in datain:
        dcmlst += [datain['nrm_ima']]

    return dcmlst


def dcmanonym(dcmpth, displayonly=False, patient='anonymised', physician='anonymised',
              dob='19800101', Cnt=None):
    '''
    Anonymise DICOM file(s)
    Arguments:
        > dcmpth:   it can be passed as a single DICOM file, or
                    a folder containing DICOM files, or a list of DICOM file paths.
        > patient:  the name of the patient.
        > physician:the name of the referring physician.
        > dob:      patient's date of birth.
        > Cnt:      dictionary of constants (containing logging variable)
    '''
    # > check if the dictionary of constant is given
    if Cnt is None:
        Cnt = {}

    # > check if a single DICOM file
    if isinstance(dcmpth, str) and os.path.isfile(dcmpth):
        dcmlst = [dcmpth]
        log.debug('recognised the input argument as a single DICOM file.')

    # > check if a folder containing DICOM files
    elif isinstance(dcmpth, str) and os.path.isdir(dcmpth):
        dircontent = os.listdir(dcmpth)
        # > create a list of DICOM files inside the folder
        dcmlst = [
            os.path.join(dcmpth, d) for d in dircontent
            if os.path.isfile(os.path.join(dcmpth, d)) and d.endswith(dcmext)]
        log.debug('recognised the input argument as the folder containing DICOM files.')

    # > check if a folder containing DICOM files
    elif isinstance(dcmpth, list):
        if not all(os.path.isfile(d) and d.endswith(dcmext) for d in dcmpth):
            raise IOError('Not all files in the list are DICOM files.')
        dcmlst = dcmpth
        log.debug('recognised the input argument as the list of DICOM file paths.')

    # > check if dictionary of data input <datain>
    elif isinstance(dcmpth, dict) and 'corepath' in dcmpth:
        dcmlst = list_dcm_datain(dcmpth)
        log.debug('recognised the input argument as the dictionary of scanner data.')

    else:
        raise IOError('Unrecognised input!')

    for dcmf in dcmlst:
        # > read the file
        dhdr = dcm.dcmread(dcmf)

        # > get the basic info about the DICOM file
        dcmtype = dcminfo(dhdr)
        log.debug(
            dedent('''\
            --------------------------------------------------
            DICOM file is for: {}
            --------------------------------------------------''').format(dcmtype))

        # > anonymise mMR data.
        if 'mmr' in dcmtype:

            if [0x029, 0x1120] in dhdr and dhdr[0x029, 0x1120].name == '[CSA Series Header Info]':
                csafield = dhdr[0x029, 0x1120]
                csa = csafield.value
            elif [0x029, 0x1020] in dhdr and dhdr[0x029,
                                                  0x1020].name == '[CSA Series Header Info]':
                csafield = dhdr[0x029, 0x1020]
                csa = csafield.value
            else:
                csa = ''

            # string length considered for replacement
            strlen = 200

            idx = [m.start() for m in re.finditer(r'([Pp]atients{0,1}[Nn]ame)', csa)]
            if idx:
                log.debug(
                    'DICOM> found sensitive information deep in the headers: {}'.format(dcmtype))

            # > run the anonymisation
            iupdate = 0

            for i in idx:
                ci = i - iupdate

                if displayonly:
                    log.debug(
                        dedent('''\
                        DICOM> sensitive info:
                             {}''').format(csa[ci:ci + strlen]))
                    continue

                rplcmnt = re.sub(r'(\{\s*\"{1,2}\W*\w+\W*\w+\W*\"{1,2}\s*\})',
                                 '{ ""' + patient + '"" }', csa[ci:ci + strlen])
                # > update string
                csa = csa[:ci] + rplcmnt + csa[ci + strlen:]
                log.debug('DICOM> removed sensitive information.')
                # > correct for the number of removed letters
                iupdate = strlen - len(rplcmnt)

            # > update DICOM
            if not displayonly and csa != '':
                csafield.value = csa

        # > Patient's name
        if [0x010, 0x010] in dhdr:
            if displayonly:
                log.debug(
                    dedent('''\
                    DICOM> sensitive info: {}
                         > {}''').format(dhdr[0x010, 0x010].name, dhdr[0x010, 0x010].value))
            else:
                dhdr[0x010, 0x010].value = patient
                log.debug('DICOM> anonymised patients name')

        # > date of birth
        if [0x010, 0x030] in dhdr:
            if displayonly:
                log.debug(
                    dedent('''\
                DICOM> sensitive info: {}
                     > {}''').format(dhdr[0x010, 0x030].name, dhdr[0x010, 0x030].value))
            else:
                dhdr[0x010, 0x030].value = dob
                log.debug('   > anonymised date of birth')

        # > physician's name
        if [0x008, 0x090] in dhdr:
            if displayonly:
                log.debug(
                    dedent('''\
                DICOM> sensitive info: {}
                     > {}''').format(dhdr[0x008, 0x090].name, dhdr[0x008, 0x090].value))
            else:
                dhdr[0x008, 0x090].value = physician
                log.debug('   > anonymised physician name')

        dhdr.save_as(dcmf)


def dcmsort(folder, copy_series=False, Cnt=None):
    '''Sort out the DICOM files in the folder according to the recorded series.'''
    # > check if the dictionary of constant is given
    if Cnt is None:
        Cnt = {}

    # list files in the input folder
    files = os.listdir(folder)

    srs = {}
    for f in files:
        if os.path.isfile(os.path.join(folder, f)) and f.endswith(dcmext):
            dhdr = dcm.read_file(os.path.join(folder, f))
            # --------------------------------
            # image size
            imsz = np.zeros(2, dtype=np.int64)
            if [0x028, 0x010] in dhdr:
                imsz[0] = dhdr[0x028, 0x010].value
            if [0x028, 0x011] in dhdr:
                imsz[1] = dhdr[0x028, 0x011].value
            # voxel size
            vxsz = np.zeros(3, dtype=np.float64)
            if [0x028, 0x030] in dhdr and [0x018, 0x050] in dhdr:
                pxsz = np.array([float(e) for e in dhdr[0x028, 0x030].value])
                vxsz[:2] = pxsz
                vxsz[2] = float(dhdr[0x018, 0x050].value)
            # orientation
            ornt = np.zeros(6, dtype=np.float64)
            if [0x020, 0x037] in dhdr:
                ornt = np.array([float(e) for e in dhdr[0x20, 0x37].value])
            # seires description, time and study time
            srs_dcrp = dhdr[0x0008, 0x103e].value
            srs_time = dhdr[0x0008, 0x0031].value[:6]
            std_time = dhdr[0x0008, 0x0030].value[:6]

            log.info(
                dedent('''\
                --------------------------------------
                DICOM series desciption: {}
                DICOM series time: {}
                DICOM study  time: {}
                --------------------------------------''').format(srs_dcrp, srs_time, std_time))

            # ---------
            # series for any category (can be multiple scans within the same category)
            recognised_series = False
            srs_k = list(srs.keys())
            for s in srs_k:
                if (np.array_equal(srs[s]['imorient'], ornt)
                        and np.array_equal(srs[s]['imsize'], imsz)
                        and np.array_equal(srs[s]['voxsize'], vxsz)
                        and srs[s]['tseries'] == srs_time):
                    recognised_series = True
                    break
            # if series was not found, create one
            if not recognised_series:
                s = srs_dcrp + '_' + srs_time
                srs[s] = {}
                srs[s]['imorient'] = ornt
                srs[s]['imsize'] = imsz
                srs[s]['voxsize'] = vxsz
                srs[s]['tseries'] = srs_time
            # append the file name
            if 'files' not in srs[s]: srs[s]['files'] = []
            if copy_series:
                srsdir = os.path.join(folder, s)
                create_dir(srsdir)
                shutil.copy(os.path.join(folder, f), srsdir)
                srs[s]['files'].append(os.path.join(srsdir, f))
            else:
                srs[s]['files'].append(os.path.join(folder, f))

    return srs


def niisort(fims, memlim=True):
    '''
    Sort all input NIfTI images and check their shape.
    Output dictionary of image files and their properties.
    Options:
        memlim -- when processing large numbers of frames the memory may
        not be large enough.  memlim causes that the output does not contain
        all the arrays corresponding to the images.
    '''
    # number of NIfTI images in folder
    Nim = 0
    # sorting list (if frame number is present in the form '_frm<dd>', where d is a digit)
    sortlist = []

    for f in fims:
        if f.endswith('.nii') or f.endswith('.nii.gz'):
            Nim += 1
            _match = re.search(r'(?<=_frm)\d*', f)
            if _match:
                frm = int(_match.group(0))
                freelists = [frm not in i for i in sortlist]
                listidx = [i for i, f in enumerate(freelists) if f]
                if listidx:
                    sortlist[listidx[0]].append(frm)
                else:
                    sortlist.append([frm])
            else:
                sortlist.append([None])

    if len(sortlist) > 1:
        # if more than one dynamic set is given, the dynamic mode is cancelled.
        dyn_flg = False
        sortlist = list(range(Nim))
    elif len(sortlist) == 1:
        dyn_flg = True
        sortlist = sortlist[0]
    else:
        raise ValueError('e> niisort input error.')

    # number of frames (can be larger than the # images)
    Nfrm = max(sortlist) + 1
    # sort the list according to the frame numbers
    _fims = ['Blank'] * Nfrm
    # list of NIfTI image shapes and data types used
    shape = []
    datype = []
    _nii = []
    for i in range(Nim):
        if dyn_flg:
            _fims[sortlist[i]] = fims[i]
            _nii = nib.load(fims[i])
            datype.append(_nii.get_data_dtype())
            shape.append(_nii.shape)
        else:
            _fims[i] = fims[i]
            _nii = nib.load(fims[i])
            datype.append(_nii.get_data_dtype())
            shape.append(_nii.shape)

    # check if all images are of the same shape and data type
    if _nii and not shape.count(_nii.shape) == len(shape):
        raise ValueError('Input images are of different shapes.')
    if _nii and not datype.count(_nii.get_data_dtype()) == len(datype):
        raise TypeError('Input images are of different data types.')
    # image shape must be 3D
    if _nii and not len(_nii.shape) == 3:
        raise ValueError('Input image(s) must be 3D.')

    out = {
        'shape': _nii.shape[::-1], 'files': _fims, 'sortlist': sortlist,
        'dtype': _nii.get_data_dtype(), 'N': Nim}

    if memlim and Nfrm > 50:
        imdic = getnii(_fims[0], output='all')
        affine = imdic['affine']
    else:
        # get the images into an array
        _imin = np.zeros((Nfrm,) + _nii.shape[::-1], dtype=_nii.get_data_dtype())
        for i in range(Nfrm):
            if i in sortlist:
                imdic = getnii(_fims[i], output='all')
                _imin[i, :, :, :] = imdic['im']
                affine = imdic['affine']
        out['im'] = _imin[:Nfrm, :, :, :]

    out['affine'] = affine

    return out


def dcm2nii(
    dcmpth,
    fimout='',
    fprefix='converted-from-DICOM_',
    fcomment='',
    outpath='',
    timestamp=True,
    executable='',
    force=False,
):
    '''
    Convert DICOM files in folder (indicated by <dcmpth>) using DCM2NIIX
    third-party software.
    '''
    # skip conversion if the output already exists and not force is selected
    if os.path.isfile(fimout) and not force:
        return fimout

    if executable == '':
        try:
            executable = rs.DCM2NIIX
        except AttributeError:
            raise AttributeError('e> could not find variable DCM2NIIX in resources.py')
    elif not os.path.isfile(executable):
        raise IOError('e> the executable is incorrect!')

    if not os.path.isdir(dcmpth):
        raise IOError('e> the provided DICOM path is not a folder!')

    # > output path
    if outpath == '' and fimout != '' and '/' in fimout:
        opth = os.path.dirname(fimout)
        if opth == '':
            opth = dcmpth
        fimout = os.path.basename(fimout)

    elif outpath == '':
        opth = dcmpth

    else:
        opth = outpath

    create_dir(opth)

    if fimout == '':
        fimout = fprefix
        if timestamp:
            fimout += time_stamp(simple_ascii=True)

    fimout = fimout.split('.nii')[0]

    # convert the DICOM mu-map images to nii
    run([executable, '-f', fimout, '-o', opth, dcmpth])

    fniiout = glob.glob(os.path.join(opth, '*' + fimout + '*.nii*'))

    if fniiout:
        return fniiout[0]
    else:
        raise ValueError('e> could not get the output file!')


def dcm2im(fpth):
    '''
    Get the DICOM files from 'fpth' into an image with the affine transformation.
    fpth can be a list of DICOM files or a path (string) to the folder with DICOM files.
    '''
    # possible DICOM file extensions
    ext = ('dcm', 'DCM', 'ima', 'IMA')

    # case when given a folder path
    if isinstance(fpth, str) and os.path.isdir(fpth):
        SZ0 = len([d for d in os.listdir(fpth) if d.endswith(ext)])
        # list of DICOM files
        fdcms = os.listdir(fpth)
        fdcms = [os.path.join(fpth, f) for f in fdcms if f.endswith(ext)]

    # case when list of DICOM files is given
    elif isinstance(fpth, list) and os.path.isfile(os.path.join(fpth[0])):
        SZ0 = len(fpth)
        # list of DICOM files
        fdcms = fpth
        fdcms = [f for f in fdcms if f.endswith(ext)]
    else:
        raise NameError('Unrecognised input for DICOM files.')

    if SZ0 < 1:
        log.error('No DICOM images in the specified path.')
        raise IOError('Input DICOM images not recognised')

    # pick single DICOM header
    dhdr = dcm.read_file(fdcms[0])

    # -----------------------------------
    # some info, e.g.: patient position and series UID
    if [0x018, 0x5100] in dhdr:
        ornt = dhdr[0x18, 0x5100].value
    else:
        ornt = 'unkonwn'
    # Series UID
    sruid = dhdr[0x0020, 0x000e].value
    # -----------------------------------

    # -----------------------------------
    # INIT
    # image position
    P = np.zeros((SZ0, 3), dtype=np.float64)
    # image orientation
    Orn = np.zeros((SZ0, 6), dtype=np.float64)
    # xy resolution
    R = np.zeros((SZ0, 2), dtype=np.float64)
    # slice thickness
    S = np.zeros((SZ0, 1), dtype=np.float64)
    # slope and intercept
    SI = np.ones((SZ0, 2), dtype=np.float64)
    SI[:, 1] = 0

    # image data as an list of array for now
    IM = []
    # -----------------------------------

    c = 0
    for d in fdcms:
        dhdr = dcm.read_file(d)
        if [0x20, 0x32] in dhdr and [0x20, 0x37] in dhdr and [0x28, 0x30] in dhdr:
            P[c, :] = np.array([float(f) for f in dhdr[0x20, 0x32].value])
            Orn[c, :] = np.array([float(f) for f in dhdr[0x20, 0x37].value])
            R[c, :] = np.array([float(f) for f in dhdr[0x28, 0x30].value])
            S[c, :] = float(dhdr[0x18, 0x50].value)
        else:
            log.error('could not read all the DICOM tags.')
            return {'im': [], 'affine': [], 'shape': [], 'orient': ornt, 'sruid': sruid}

        if [0x28, 0x1053] in dhdr and [0x28, 0x1052] in dhdr:
            SI[c, 0] = float(dhdr[0x28, 0x1053].value)
            SI[c, 1] = float(dhdr[0x28, 0x1052].value)
        IM.append(dhdr.pixel_array)
        c += 1

    # check if orientation/resolution is the same for all slices
    if np.sum(Orn - Orn[0, :]) > 1e-6:
        log.error('varying orientation for slices')
    else:
        Orn = Orn[0, :]
    if np.sum(R - R[0, :]) > 1e-6:
        log.error('varying resolution for slices')
    else:
        R = R[0, :]

    # Patient Position
    # Rows and Columns
    if [0x28, 0x10] in dhdr and [0x28, 0x11] in dhdr:
        SZ2 = dhdr[0x28, 0x10].value
        SZ1 = dhdr[0x28, 0x11].value
    # image resolution
    SZ_VX2 = R[0]
    SZ_VX1 = R[1]

    # now sort the images along k-dimension
    k = np.argmin(abs(Orn[:3] + Orn[3:]))
    # sorted indeces
    si = np.argsort(P[:, k])
    Pos = np.zeros(P.shape, dtype=np.float64)
    im = np.zeros((SZ0, SZ1, SZ2), dtype=np.float32)

    # check if the detentions are in agreement (the pixel array could be transposed...)
    if IM[0].shape[0] == SZ1:
        for i in range(SZ0):
            im[i, :, :] = IM[si[i]] * SI[si[i], 0] + SI[si[i], 1]
            Pos[i, :] = P[si[i]]
    else:
        for i in range(SZ0):
            im[i, :, :] = IM[si[i]].T * SI[si[i], 0] + SI[si[i], 1]
            Pos[i, :] = P[si[i]]

    # proper slice thickness
    Zz = (P[si[-1], 2] - P[si[0], 2]) / (SZ0-1)
    Zy = (P[si[-1], 1] - P[si[0], 1]) / (SZ0-1)
    Zx = (P[si[-1], 0] - P[si[0], 0]) / (SZ0-1)

    # dictionary for affine and image size for the image
    A = {
        'AFFINE': np.array([[SZ_VX2 * Orn[0], SZ_VX1 * Orn[3], Zx, Pos[0, 0]],
                            [SZ_VX2 * Orn[1], SZ_VX1 * Orn[4], Zy, Pos[0, 1]],
                            [SZ_VX2 * Orn[2], SZ_VX1 * Orn[5], Zz, Pos[0, 2]], [0., 0., 0., 1.]]),
        'SHAPE': (SZ0, SZ1, SZ2)}

    # the returned image is already scaled according to the dcm header
    return {'im': im, 'affine': A['AFFINE'], 'shape': A['SHAPE'], 'orient': ornt, 'sruid': sruid}



#!/usr/bin/env python3
"""
Compile CUDA source code and setup Python 3 package 'nimpa'
for namespace 'niftypet'.
"""
import logging
import re
import sys
from pathlib import Path
from textwrap import dedent

from setuptools import find_packages, setup
from setuptools_scm import get_version

from niftypet.ninst import cudasetup as cs
from niftypet.ninst import dinf
from niftypet.ninst import install_tools as tls

__version__ = get_version(root=".", relative_to=__file__)

logging.basicConfig(level=logging.INFO, format=tls.LOG_FORMAT)
log = logging.getLogger("nimpa.setup")

tls.check_platform()
ext = tls.check_depends() # external dependencies

if not ext["git"]:
    raise SystemError(
        dedent("""\
            --------------------------------------------------------------
            Git is not installed but is required for tools installation.
            --------------------------------------------------------------"""))

cs.resources_setup(gpu=False) # install resources.py
try:
    gpuarch = cs.dev_setup()  # update resources.py with a supported GPU device
except Exception as exc:
    log.error("could not set up CUDA:\n%s", exc)
    gpuarch = None

# First install third party apps for NiftyPET tools
log.info(
    dedent("""\
        --------------------------------------------------------------
        Setting up NiftyPET tools ...
        --------------------------------------------------------------"""))
# get the local path to NiftyPET resources.py
path_resources = cs.path_niftypet_local()
# if exists, import the resources and get the constants
resources = cs.get_resources()
# get the current setup, if any
Cnt = resources.get_setup()
# check the installation of tools
chck_tls = tls.check_version(Cnt, chcklst=["RESPATH", "REGPATH", "DCM2NIIX"])

# -------------------------------------------
# NiftyPET tools:
# -------------------------------------------
if "sdist" not in sys.argv or any(i in sys.argv for i in ["build", "bdist", "wheel"]):
    # DCM2NIIX
    if not chck_tls["DCM2NIIX"]:
        # reply = tls.query_yesno("q> dcm2niix not found.\n   Install it?")
        # if reply:
        log.info(
            dedent("""\
            --------------------------------------------------------------
            Installing dcm2niix ...
            --------------------------------------------------------------"""))
        Cnt = tls.install_tool("dcm2niix", Cnt)
    # ---------------------------------------

    # -------------------------------------------
    # NiftyReg
    if not chck_tls["REGPATH"] or not chck_tls["RESPATH"]:
        reply = True
        if not gpuarch:
            try:
                reply = tls.query_yesno(
                    "q> the latest compatible version of NiftyReg seems to be missing.\n"
                    "   Do you want to install it?")
            except BaseException:
                pass

        if reply:
            log.info(
                dedent("""\
                    --------------------------------------------------------------
                    Installing NiftyReg ...
                    --------------------------------------------------------------"""))
            Cnt = tls.install_tool("niftyreg", Cnt)
    log.info(
        dedent("""\
            --------------------------------------------------------------
            Installation of NiftyPET-tools is done.
            --------------------------------------------------------------"""))
else:
    log.info(
        dedent("""\
            --------------------------------------------------------------
            Skipping installation of NiftyPET-tools.
            --------------------------------------------------------------"""))

build_ver = ".".join(__version__.split('.')[:3]).split(".dev")[0]
setup_kwargs = {
    "use_scm_version": True, "packages": find_packages(exclude=["tests"]),
    "package_data": {"niftypet": ["nimpa/auxdata/*"]}}
cmake_args = [
    f"-DNIMPA_BUILD_VERSION={build_ver}", f"-DPython3_ROOT_DIR={sys.prefix}",
    f"-DNIMPA_KERNEL_RADIUS={getattr(resources, 'RSZ_PSF_KRNL', 8)}"]

try:
    from skbuild import setup as sksetup
    nvcc_arches = {"{2:d}{3:d}".format(*i) for i in dinf.gpuinfo() if i[2:4] >= (3, 5)}
    if nvcc_arches:
        cmake_args.append("-DCMAKE_CUDA_ARCHITECTURES=" + ";".join(sorted(nvcc_arches)))
except Exception as exc:
    log.warning("Import or CUDA device detection error:\n%s", exc)
    setup(**setup_kwargs)
else:
    for i in (Path(__file__).resolve().parent / "_skbuild").rglob("CMakeCache.txt"):
        i.write_text(re.sub("^//.*$\n^[^#].*pip-build-env.*$", "", i.read_text(), flags=re.M))
    sksetup(cmake_source_dir="niftypet", cmake_languages=("C", "CXX", "CUDA"),
            cmake_minimum_required_version="3.18", cmake_args=cmake_args, **setup_kwargs)
